# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Analysis class to characterize correlated readout error
"""
from typing import List, Tuple
import numpy as np
import matplotlib.pyplot as plt
from qiskit.result import CorrelatedReadoutMitigator
from qiskit_experiments.framework import ExperimentData
from qiskit_experiments.framework.matplotlib import get_non_gui_ax
from qiskit_experiments.framework import BaseAnalysis, AnalysisResultData, Options


class CorrelatedReadoutErrorAnalysis(BaseAnalysis):
    r"""An analysis to characterize correlated readout error.

    # section: overview

        This class generates the full assignment matrix :math:`A` characterizing the
        readout error for the given qubits from the experiment results
        and returns the resulting :class:`~qiskit.result.CorrelatedReadoutMitigator`

        :math:`A` is a :math:`2^n\times 2^n` matrix :math:`A` such that :math:`A_{y,x}`
        is the probability to observe :math:`y` given the true outcome should be :math:`x`.

        In the experiment, for each :math:`x` a circuit is constructed whose expected
        outcome is :math:`x`. From the observed results on the circuit, the probability for
        each :math:`y` is determined, and :math:`A_{y,x}` is set accordingly.

        Analysis Results:
           * "Local Readout Mitigator": The :class:`~qiskit.result.LocalReadoutMitigator`.

        Analysis Figures:
           * (Optional) A figure of the assignment matrix.

    # section: reference
        .. ref_arxiv:: 1 2006.14044
    """

    @classmethod
    def _default_options(cls) -> Options:
        """Return default analysis options.

        Analysis Options:
            plot (bool): Set ``True`` to create figure for fit result.
            ax (AxesSubplot): Optional. A matplotlib axis object to draw.
        """
        options = super()._default_options()
        options.plot = True
        options.ax = None
        return options

    def _run_analysis(
        self, experiment_data: ExperimentData, **options
    ) -> Tuple[List[AnalysisResultData], List["matplotlib.figure.Figure"]]:
        data = experiment_data.data()
        qubits = experiment_data.metadata["physical_qubits"]
        labels = [datum["metadata"]["state_label"] for datum in data]
        matrix = self._generate_matrix(data, labels)
        result_mitigator = CorrelatedReadoutMitigator(matrix, qubits=qubits)
        analysis_results = [AnalysisResultData("Correlated Readout Mitigator", result_mitigator)]
        if self.options.plot:
            ax = options.get("ax", None)
            figures = [self._assignment_matrix_visualization(matrix, labels, ax)]
        else:
            figures = []
        return analysis_results, figures

    def _generate_matrix(self, data, labels) -> np.array:
        list_size = len(labels)
        matrix = np.zeros([list_size, list_size], dtype=float)
        # matrix[i][j] is the probability of counting i for expected j
        for datum in data:
            expected_outcome = datum["metadata"]["state_label"]
            j = labels.index(expected_outcome)
            total_counts = sum(datum["counts"].values())
            for measured_outcome, count in datum["counts"].items():
                i = labels.index(measured_outcome)
                matrix[i][j] = count / total_counts
        return matrix

    def _assignment_matrix_visualization(
        self, matrix, labels, ax=None
    ) -> "matplotlib.figure.Figure":
        """
        Plot the assignment matrix (2D color grid plot).

        Args:
            matrix: assignment matrix to plot
            ax (matplotlib.axes): settings for the graph

        Returns:
            The generated plot of the assignment matrix

        Raises:
            QiskitError: If _cal_matrices was not set.

            ImportError: If matplotlib was not installed.

        """

        if ax is None:
            ax = get_non_gui_ax()
        figure = ax.get_figure()
        ax.matshow(matrix, cmap=plt.cm.binary, clim=[0, 1])
        ax.set_xlabel("Prepared State")
        ax.xaxis.set_label_position("top")
        ax.set_ylabel("Measured State")
        ax.set_xticks(np.arange(len(labels)))
        ax.set_yticks(np.arange(len(labels)))
        ax.set_xticklabels(labels)
        ax.set_yticklabels(labels)
        return figure



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Plotter for curve fits, specifically from :class:`.CurveAnalysis`."""
from typing import List

from uncertainties import UFloat

from qiskit_experiments.curve_analysis.utils import analysis_result_to_repr
from qiskit_experiments.framework import Options

from .base_plotter import BasePlotter


class CurvePlotter(BasePlotter):
    """A plotter class to plot results from :class:`.CurveAnalysis`.

    ``CurvePlotter`` plots results from curve fits, which includes

        - Raw results as a scatter plot.
        - Processed results with standard deviations/confidence intervals.
        - Interpolated fit results from the curve analysis.
        - Confidence interval for the fit results.
        - A report on the performance of the fit.
    """

    @classmethod
    def expected_series_data_keys(cls) -> List[str]:
        """Returns the expected series data keys supported by this plotter.

        Data Keys:
            x: X-values for raw results.
            y: Y-values for raw results. Goes with ``x``.
            x_formatted: X-values for processed results.
            y_formatted: Y-values for processed results. Goes with ``x_formatted``.
            y_formatted_err: Error in ``y_formatted``, to be plotted as error-bars.
            x_interp: Interpolated X-values for a curve fit.
            y_interp: Y-values corresponding to the fit for ``y_interp`` X-values.
            y_interp_err: The standard deviations of the fit for each X-value in
                ``y_interp``. This data key relates to the option ``plot_sigma``.
            x_residuals: The X-values for the residual plot.
            y_residuals: The residual from the fitting.
        """
        return [
            "x",
            "y",
            "x_formatted",
            "y_formatted",
            "y_formatted_err",
            "x_interp",
            "y_interp",
            "y_interp_err",
            "x_residuals",
            "y_residuals",
        ]

    @classmethod
    def expected_supplementary_data_keys(cls) -> List[str]:
        """Returns the expected figures data keys supported by this plotter.

        This plotter generates a single text box, i.e. fit report, by digesting the
        provided supplementary data. The style and position of the report is controlled
        by ``textbox_rel_pos`` and ``textbox_text_size`` style parameters in
        :class:`PlotStyle`.

        Data Keys:
            primary_results: A list of :class:`.AnalysisResultData` objects to be shown
                in the fit report window. Typically, these are fit parameter values or
                secondary quantities computed from multiple fit parameters.
            fit_red_chi: The best reduced-chi squared value of the fit curves. If the
                fit consists of multiple sub-fits, this will be a dictionary keyed on
                the analysis name. Otherwise, this is a single float value of a
                particular analysis.
        """
        return [
            "primary_results",
            "fit_red_chi",
        ]

    @classmethod
    def _default_options(cls) -> Options:
        """Return curve-plotter specific default plotter options.

        Options:
            plot_sigma (List[Tuple[float, float]]): A list of two number tuples showing
                the configuration to write confidence intervals for the fit curve. The
                first argument is the relative sigma (n_sigma), and the second argument
                is the transparency of the interval plot in ``[0, 1]``. Multiple n_sigma
                intervals can be drawn for the same curve.

        """
        options = super()._default_options()
        options.plot_sigma = [(1.0, 0.7), (3.0, 0.3)]
        return options

    @classmethod
    def _default_figure_options(cls) -> Options:
        r"""Return curve-plotter specific default figure options.

        Figure Options:
            report_red_chi2_label (str): The label for the reduced-chi squared entry of
                the fit report. Defaults to the Python string literal
                ``"reduced-$\\chi^2$"``, corresponding to the formatted string
                reduced-:math:`\chi^2`.
        """
        fig_opts = super()._default_figure_options()
        fig_opts.report_red_chi2_label = "reduced-$\\chi^2$"

        return fig_opts

    def _plot_figure(self):
        """Plots a curve fit figure."""
        for ser in self.series:
            # Scatter plot with error-bars
            plotted_formatted_data = False
            if self.data_exists_for(ser, ["x_formatted", "y_formatted", "y_formatted_err"]):
                x, y, yerr = self.data_for(ser, ["x_formatted", "y_formatted", "y_formatted_err"])
                self.drawer.scatter(x, y, y_err=yerr, name=ser, zorder=2, legend=True)
                plotted_formatted_data = True

            # Scatter plot
            if self.data_exists_for(ser, ["x", "y"]):
                x, y = self.data_for(ser, ["x", "y"])
                options = {
                    "zorder": 1,
                }
                # If we plotted formatted data, differentiate scatter points by setting normal X-Y
                # markers to gray.
                if plotted_formatted_data:
                    options["color"] = "gray"
                # If we didn't plot formatted data, the X-Y markers should be used for the legend. We add
                # it to ``options`` so it's easier to pass to ``scatter``.
                if not plotted_formatted_data:
                    options["legend"] = True
                self.drawer.scatter(
                    x,
                    y,
                    name=ser,
                    **options,
                )

            # Line plot for fit
            if self.data_exists_for(ser, ["x_interp", "y_interp"]):
                x, y = self.data_for(ser, ["x_interp", "y_interp"])
                self.drawer.line(x, y, name=ser, zorder=3)

            # Confidence interval plot
            if self.data_exists_for(ser, ["x_interp", "y_interp", "y_interp_err"]):
                x, y_interp, y_interp_err = self.data_for(
                    ser, ["x_interp", "y_interp", "y_interp_err"]
                )
                for n_sigma, alpha in self.options.plot_sigma:
                    self.drawer.filled_y_area(
                        x,
                        y_interp + n_sigma * y_interp_err,
                        y_interp - n_sigma * y_interp_err,
                        name=ser,
                        alpha=alpha,
                        zorder=5,
                    )

            # Plot residuals
            if self.data_exists_for(ser, ["x_residuals", "y_residuals"]):
                # check if we cancel residuals plotting
                if self.options.get("style", {}).get("style_name") != "canceled_residuals":
                    series_name = ser + "_residuals"
                    x, y = self.data_for(ser, ["x_residuals", "y_residuals"])
                    self.drawer.scatter(x, y, name=series_name, legend=True)

            # Fit report
            report = self._write_report()
            if len(report) > 0:
                self.drawer.textbox(report)

    def _write_report(self) -> str:
        """Write fit report with supplementary_data.

        Subclass can override this method to customize fit report. By default, this
        writes important fit parameters and chi-squared value of the fit in the fit
        report. The ``report_red_chi2_label`` figure option controls the label for the
        chi-squared entries in the report.

        Returns:
            Fit report.
        """
        report = ""

        if "primary_results" in self.supplementary_data:
            lines = []
            for outcome in self.supplementary_data["primary_results"]:
                if isinstance(outcome.value, (float, UFloat)):
                    lines.append(analysis_result_to_repr(outcome))
            report += "\n".join(lines)

        if "fit_red_chi" in self.supplementary_data:
            red_chi = self.supplementary_data["fit_red_chi"]
            if len(report) > 0:
                report += "\n"
            if isinstance(red_chi, float):
                report += f"{self.figure_options.report_red_chi2_label} = {red_chi: .4g}"
            else:
                # Composite curve analysis reporting multiple chi-sq values.
                # This is usually given by a dict keyed on fit group name.

                # Add gap between primary-results and reduced-chi squared as
                # we have multiple values to display. This is easier to read.
                if len(report) > 0:
                    report += "\n"

                # Created indented text of reduced-chi squared results.
                report += f"{self.figure_options.report_red_chi2_label} per fit\n"
                lines = []
                for mod_name, mod_chi in red_chi.items():
                    lines.append(f"  * {mod_name}: {mod_chi: .4g}")
                report += "\n".join(lines)

        return report



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Discriminator wrappers to make discriminators serializable.."""

from abc import abstractmethod
from typing import Any, Dict, List


class BaseDiscriminator:
    """An abstract base class for serializable discriminators used in the
    :class:`.DiscriminatorNode` data action nodes.

    This class allows developers to implement their own discriminators or wrap discriminators
    from external libraries which therefore ensures that the discriminator fits in
    the data processing chain. This class defines an interface for discriminator objects.
    Subclasses must implement the following methods:

    - :meth:`predict`: called in the :class:`.Discriminator` data-action class to predict
      labels from the input level-one data.
    - :meth:`config`: produces the config file to serialize and deserialize the discriminator.
    - :meth:`is_trained`: indicates if the discriminator is trained, i.e., fit to training data.
    """

    @abstractmethod
    def predict(self, data: List):
        """The function used to predict the labels of the data."""

    @property
    def discriminator(self) -> Any:
        """Return the discriminator object that is wrapped.

        Sub-classes may not need to implement this method but can chose to
        if they are wrapping an object capable of discrimination.
        """
        return None

    @abstractmethod
    def config(self) -> Dict[str, Any]:
        """Return the configuration of the discriminator."""

    @abstractmethod
    def is_trained(self) -> bool:
        """Return True if this discriminator has been trained on data."""

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "BaseDiscriminator":
        """Create a discriminator from the configuration."""

    def __json_encode__(self):
        """Convert to format that can be JSON serialized."""
        return self.config()

    @classmethod
    def __json_decode__(cls, value: Dict[str, Any]) -> "BaseDiscriminator":
        """Load from JSON compatible format."""
        return cls.from_config(value)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Error amplification analysis."""

from typing import List, Union, Optional

import lmfit
import numpy as np

import qiskit_experiments.curve_analysis as curve


class ErrorAmplificationAnalysis(curve.CurveAnalysis):
    r"""Error amplification analysis class based on a fit to a cosine function.

    # section: fit_model

        Analyse an error amplifying calibration experiment by fitting the data to a cosine
        function. The user must also specify the intended rotation angle per gate, here labeled,
        :math:`{\rm apg}`. The parameter of interest in the fit is the deviation from the
        intended rotation angle per gate labeled :math:`{\rm d}\theta`. The fit function is

        .. math::
            y = \frac{{\rm amp}}{2}\cos\left(x[{\rm d}\theta + {\rm apg} ] \
            -{\rm phase\_offset}\right)+{\rm base}

        To understand how the error is measured we can transformed the function above into

        .. math::
            y = \frac{{\rm amp}}{2} \left(\
            \cos\right({\rm d}\theta \cdot x\left)\
            \cos\right({\rm apg} \cdot x - {\rm phase\_offset}\left) -\
            \sin\right({\rm d}\theta \cdot x\left)\
            \sin\right({\rm apg} \cdot x - {\rm phase\_offset}\left)
            \right) + {\rm base}

        When :math:`{\rm apg} \cdot x - {\rm phase\_offset} = (2n + 1) \pi/2` is satisfied the
        fit model above simplifies to

        .. math::
            y = \mp \frac{{\rm amp}}{2} \sin\left({\rm d}\theta \cdot x\right) + {\rm base}

        In the limit :math:`{\rm d}\theta \ll 1`, the error can be estimated from the curve data

        .. math::
            {\rm d}\theta \simeq \mp \frac{2(y - {\rm base})}{x \cdot {\rm amp}}


    # section: fit_parameters
        defpar \rm amp:
            desc: Amplitude of the oscillation.
            init_guess: The maximum y value less the minimum y value.
            bounds: [-2, 2] scaled to the maximum signal value.

        defpar \rm base:
            desc: Base line.
            init_guess: The average of the data.
            bounds: [-1, 1] scaled to the maximum signal value.

        defpar d\theta:
            desc: The angle offset in the gate that we wish to measure.
            init_guess: Multiple initial guesses are tried ranging from -a to a
                where a is given by :code:`max(abs(angle_per_gate), np.pi / 2)`.
                Extra guesses are added based on curve data when either :math:`\rm amp` or
                :math:`\rm base` is :math:`\pi/2`. See fit model for details.
            bounds: [-0.8 pi, 0.8 pi]. The bounds do not include plus and minus pi since these values
                often correspond to symmetry points of the fit function. Furthermore,
                this type of analysis is intended for values of :math:`d\theta` close to zero.

    """

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        super().__init__(
            models=[
                lmfit.models.ExpressionModel(
                    expr="amp / 2 * cos((d_theta + angle_per_gate) * x - phase_offset) + base",
                    name="ping_pong",
                )
            ],
            name=name,
        )

    @classmethod
    def _default_options(cls):
        r"""Return the default analysis options.

        See :meth:`~qiskit_experiment.curve_analysis.CurveAnalysis._default_options` for
        descriptions of analysis options.

        Analysis Options:
            max_good_angle_error (float): The maximum angle error for which the fit is
                considered as good. Defaults to :math:`\pi/2`.
        """
        default_options = super()._default_options()
        default_options.plotter.set_figure_options(
            xlabel="Number of gates (n)",
            ylabel="Population",
            ylim=(0, 1.0),
        )
        default_options.result_parameters = ["d_theta"]
        default_options.max_good_angle_error = np.pi / 2

        return default_options

    def _generate_fit_guesses(
        self,
        user_opt: curve.FitOptions,
        curve_data: curve.ScatterTable,
    ) -> Union[curve.FitOptions, List[curve.FitOptions]]:
        """Create algorithmic initial fit guess from analysis options and curve data.

        Args:
            user_opt: Fit options filled with user provided guess and bounds.
            curve_data: Formatted data collection to fit.

        Returns:
            List of fit options that are passed to the fitter function.
        """
        fixed_params = self.options.fixed_parameters

        max_abs_y, _ = curve.guess.max_height(curve_data.y, absolute=True)
        max_y, min_y = np.max(curve_data.y), np.min(curve_data.y)

        user_opt.bounds.set_if_empty(
            d_theta=(-0.8 * np.pi, 0.8 * np.pi), base=(-max_abs_y, max_abs_y)
        )
        user_opt.p0.set_if_empty(base=(max_y + min_y) / 2)

        if "amp" in user_opt.p0:
            user_opt.p0.set_if_empty(amp=max_y - min_y)
            user_opt.bounds.set_if_empty(amp=(0, 2 * max_abs_y))
            amp = user_opt.p0["amp"]
        else:
            # Fixed parameter
            amp = fixed_params.get("amp", 1.0)

        # Base the initial guess on the intended angle_per_gate and phase offset.
        apg = user_opt.p0.get("angle_per_gate", fixed_params.get("angle_per_gate", 0.0))
        phi = user_opt.p0.get("phase_offset", fixed_params.get("phase_offset", 0.0))

        # Prepare logical guess for specific condition (often satisfied)
        d_theta_guesses = []

        offsets = apg * curve_data.x + phi
        for i in range(curve_data.x.size):
            xi = curve_data.x[i]
            yi = curve_data.y[i]
            if np.isclose(offsets[i] % np.pi, np.pi / 2) and xi > 0:
                # Condition satisfied: i.e. cos(apg x - phi) = 0
                err = -np.sign(np.sin(offsets[i])) * (yi - user_opt.p0["base"]) / (0.5 * amp)
                # Validate estimate. This is just the first order term of Maclaurin expansion.
                if np.abs(err) < 0.5:
                    d_theta_guesses.append(err / xi)
                else:
                    # Terminate guess generation because larger d_theta x will start to
                    # reduce net y value and underestimate the rotation.
                    break

        # Add naive guess for more coverage
        guess_range = max(abs(apg), np.pi / 2)
        d_theta_guesses.extend(np.linspace(-guess_range, guess_range, 11))

        options = []
        for d_theta_guess in d_theta_guesses:
            new_opt = user_opt.copy()
            new_opt.p0.set_if_empty(d_theta=d_theta_guess)
            options.append(new_opt)

        return options

    def _evaluate_quality(self, fit_data: curve.CurveFitResult) -> Union[str, None]:
        """Algorithmic criteria for whether the fit is good or bad.

        A good fit has:
            - a reduced chi-squared lower than three and greater than zero,
            - a measured angle error that is smaller than the allowed maximum good angle error.
              This quantity is set in the analysis options.
        """
        fit_d_theta = fit_data.ufloat_params["d_theta"]

        criteria = [
            0 < fit_data.reduced_chisq < 3,
            abs(fit_d_theta.nominal_value) < abs(self.options.max_good_angle_error),
        ]

        if all(criteria):
            return "good"

        return "bad"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Linear inversion MLEtomography fitter.
"""

from typing import Dict, Tuple, Optional, Sequence, List
from functools import lru_cache
import time
import numpy as np
from qiskit_experiments.library.tomography.basis import (
    MeasurementBasis,
    PreparationBasis,
    LocalMeasurementBasis,
    LocalPreparationBasis,
)
from .lstsq_utils import _partial_outcome_function
from .fitter_data import _basis_dimensions


def linear_inversion(
    outcome_data: np.ndarray,
    shot_data: np.ndarray,
    measurement_data: np.ndarray,
    preparation_data: np.ndarray,
    measurement_basis: Optional[MeasurementBasis] = None,
    preparation_basis: Optional[PreparationBasis] = None,
    measurement_qubits: Optional[Tuple[int, ...]] = None,
    preparation_qubits: Optional[Tuple[int, ...]] = None,
    conditional_measurement_indices: Optional[np.ndarray] = None,
    conditional_preparation_indices: Optional[np.ndarray] = None,
    atol: float = 1e-8,
) -> Tuple[np.ndarray, Dict]:
    r"""Linear inversion tomography fitter.

    Overview
        This fitter uses linear inversion to reconstructs the maximum-likelihood
        estimate of the least-squares log-likelihood function

        .. math::
            \hat{\rho}
                &= -\mbox{argmin }\log\mathcal{L}{\rho} \\
                &= \mbox{argmin }\sum_i (\mbox{Tr}[E_j\rho] - \hat{p}_i)^2 \\
                &= \mbox{argmin }\|Ax - y \|_2^2

        where

        * :math:`A = \sum_j |j \rangle\!\langle\!\langle E_j|` is the matrix of measured
          basis elements.
        * :math:`y = \sum_j \hat{p}_j |j\rangle` is the vector of estimated measurement
          outcome probabilities for each basis element.
        * :math:`x = |\rho\rangle\!\rangle` is the vectorized density matrix.

    Additional Details
        The linear inversion solution is given by

        .. math::
            \hat{\rho} = \sum_i \hat{p}_i D_i

        where measurement probabilities :math:`\hat{p}_i = f_i / n_i` are estimated
        from the observed count frequencies :math:`f_i` in :math:`n_i` shots for each
        basis element :math:`i`, and :math:`D_i` is the *dual basis* element constructed
        from basis :math:`\{E_i\}` via:

        .. math:

            |D_i\rangle\!\rangle = M^{-1}|E_i \rangle\!\rangle \\
            M = \sum_j |E_j\rangle\!\rangle\!\langle\!\langle E_j|

    .. note::

        The Linear inversion fitter treats the input measurement and preparation
        bases as local bases and constructs separate 1-qubit dual basis for each
        individual qubit.

        Linear inversion is only possible if the input bases are local and a spanning
        set for the vector space of the reconstructed matrix
        (*tomographically complete*). If the basis is not tomographically complete
        the :func:`~qiskit_experiments.library.tomography.fitters.scipy_linear_lstsq`
        or :func:`~qiskit_experiments.library.tomography.fitters.cvxpy_linear_lstsq`
        function can be used to solve the same objective function via
        least-squares optimization.

    Args:
        outcome_data: basis outcome frequency data.
        shot_data: basis outcome total shot data.
        measurement_data: measurement basis index data.
        preparation_data: preparation basis index data.
        measurement_basis: the tomography measurement basis.
        preparation_basis: the tomography preparation basis.
        measurement_qubits: Optional, the physical qubits that were measured.
             If None they are assumed to be [0, ..., M-1] for M measured qubits.
        preparation_qubits: Optional, the physical qubits that were prepared.
            If None they are assumed to be [0, ..., N-1] forN prepared qubits.
        conditional_measurement_indices: Optional, conditional measurement data
            indices. If set this will return a list of fitted states conditioned
            on a fixed basis measurement of these qubits.
        conditional_preparation_indices: Optional, conditional preparation data
            indices. If set this will return a list of fitted states conditioned
            on a fixed basis preparation of these qubits.
        atol: truncate any probabilities below this value to zero.

    Raises:
        AnalysisError: If the fitted vector is not a square matrix

    Returns:
        The fitted matrix rho.
    """
    t_start = time.time()

    if measurement_basis and measurement_qubits is None:
        measurement_qubits = tuple(range(measurement_data.shape[1]))
    if preparation_basis and preparation_qubits is None:
        preparation_qubits = tuple(range(preparation_data.shape[1]))

    input_dims = _basis_dimensions(
        basis=preparation_basis,
        qubits=preparation_qubits,
        conditional_indices=conditional_preparation_indices,
    )
    output_dims = _basis_dimensions(
        basis=measurement_basis,
        qubits=measurement_qubits,
        conditional_indices=conditional_measurement_indices,
    )

    metadata = {
        "fitter": "linear_inversion",
        "input_dims": input_dims,
        "output_dims": output_dims,
    }

    if conditional_preparation_indices:
        # Split measurement qubits into conditional and non-conditional qubits
        f_prep_qubits = []
        prep_indices = []
        for i, qubit in enumerate(preparation_qubits):
            if i not in conditional_preparation_indices:
                f_prep_qubits.append(qubit)
                prep_indices.append(i)
        # Indexing array for fully tomo measured qubits
        f_prep_indices = np.array(prep_indices, dtype=int)
        f_cond_prep_indices = np.array(conditional_preparation_indices, dtype=int)
    else:
        f_prep_qubits = preparation_qubits
        f_prep_indices = slice(None)
        f_cond_prep_indices = slice(0, 0)

    if conditional_measurement_indices:
        # Split measurement qubits into conditional and non-conditional qubits
        f_cond_meas_qubits = []
        f_meas_qubits = []
        meas_indices = []
        for i, qubit in enumerate(measurement_qubits):
            if i in conditional_measurement_indices:
                f_cond_meas_qubits.append(qubit)
            else:
                f_meas_qubits.append(qubit)
                meas_indices.append(i)

        cond_meas_size = np.prod(measurement_basis.outcome_shape(f_cond_meas_qubits), dtype=int)

        # Indexing array for fully tomo measured qubits
        f_meas_indices = np.array(meas_indices, dtype=int)
        f_cond_meas_indices = np.array(conditional_measurement_indices, dtype=int)

        # Reduced outcome functions
        f_meas_outcome = _partial_outcome_function(tuple(f_meas_indices))
        f_cond_meas_outcome = _partial_outcome_function(tuple(conditional_measurement_indices))
    else:
        cond_meas_size = 1
        f_meas_qubits = measurement_qubits
        f_meas_indices = slice(None)
        f_cond_meas_indices = slice(0, 0)

        def f_meas_outcome(x):
            return x

        def f_cond_meas_outcome(_):
            return 0

    # Construct dual bases
    meas_dual_basis = None
    if measurement_basis and f_meas_qubits:
        meas_duals = {i: _dual_povms(measurement_basis, i) for i in f_meas_qubits}
        meas_dual_basis = LocalMeasurementBasis(
            f"Dual_{measurement_basis.name}", qubit_povms=meas_duals
        )

    prep_dual_basis = None
    if preparation_basis and f_prep_qubits:
        prep_duals = {i: _dual_states(preparation_basis, i) for i in f_prep_qubits}
        prep_dual_basis = LocalPreparationBasis(
            f"Dual_{preparation_basis.name}", qubit_states=prep_duals
        )

    if shot_data is None:
        # Define shots by sum of all outcome frequencies for each basis
        shot_data = np.sum(outcome_data, axis=(0, -1))

    # Calculate shape of matrix to be fitted
    if prep_dual_basis:
        pdim = np.prod(prep_dual_basis.matrix_shape(f_prep_qubits), dtype=int)
    else:
        pdim = 1
    if meas_dual_basis:
        mdim = np.prod(meas_dual_basis.matrix_shape(f_meas_qubits), dtype=int)
    else:
        mdim = 1
    shape = (pdim * mdim, pdim * mdim)

    # Construct linear inversion matrix
    cond_circ_size = outcome_data.shape[0]
    cond_fits = []
    if cond_circ_size > 1:
        metadata["conditional_circuit_outcome"] = []
    if conditional_measurement_indices:
        metadata["conditional_measurement_index"] = []
        metadata["conditional_measurement_outcome"] = []
    if conditional_preparation_indices:
        metadata["conditional_preparation_index"] = []
    for circ_idx in range(cond_circ_size):
        fits = {}
        for i, outcomes in enumerate(outcome_data[circ_idx]):
            shots = shot_data[i]
            pidx = preparation_data[i][f_prep_indices]
            midx = measurement_data[i][f_meas_indices]
            cond_prep_key = tuple(preparation_data[i][f_cond_prep_indices])
            cond_meas_key = tuple(measurement_data[i][f_cond_meas_indices])
            key = (cond_prep_key, cond_meas_key)
            if key not in fits:
                fits[key] = [np.zeros(shape, dtype=complex) for _ in range(cond_meas_size)]

            # Get prep basis component
            if prep_dual_basis:
                p_mat = np.transpose(prep_dual_basis.matrix(pidx, f_prep_qubits))
            else:
                p_mat = 1

            # Get probabilities and optional measurement basis component
            for outcome, freq in enumerate(outcomes):
                prob = freq / shots
                if np.isclose(prob, 0, atol=atol):
                    # Skip component with zero probability
                    continue

                # Get component on non-conditional bits
                outcome_meas = f_meas_outcome(outcome)
                if meas_dual_basis:
                    dual_op = meas_dual_basis.matrix(midx, outcome_meas, f_meas_qubits)
                    if prep_dual_basis:
                        dual_op = np.kron(p_mat, dual_op)
                else:
                    dual_op = p_mat

                # Add component to correct conditional
                outcome_cond = f_cond_meas_outcome(outcome)
                fits[key][outcome_cond] += prob * dual_op

        # Append conditional fit metadata
        for (prep_key, meas_key), c_fits in fits.items():
            cond_fits += c_fits
            if cond_circ_size > 1:
                metadata["conditional_circuit_outcome"] += len(c_fits) * [circ_idx]
            if conditional_measurement_indices:
                metadata["conditional_measurement_index"] += len(c_fits) * [meas_key]
                metadata["conditional_measurement_outcome"] += list(range(cond_meas_size))
            if conditional_preparation_indices:
                metadata["conditional_preparation_index"] += len(c_fits) * [prep_key]

    t_stop = time.time()
    metadata["fitter_time"] = t_stop - t_start

    if len(cond_fits) == 1:
        return cond_fits[0], metadata
    return cond_fits, metadata


@lru_cache(None)
def _dual_states(basis: PreparationBasis, qubit: int) -> np.ndarray:
    """Construct a dual preparation basis for linear inversion"""
    size = basis.index_shape((qubit,))[0]
    states = np.asarray([basis.matrix((i,), (qubit,)) for i in range(size)])
    return _construct_dual_states(states)


@lru_cache(None)
def _dual_povms(basis: MeasurementBasis, qubit: int) -> List[List[np.ndarray]]:
    """Construct dual POVM states for linear inversion"""
    size = basis.index_shape((qubit,))[0]
    num_outcomes = basis.outcome_shape((qubit,))[0]

    # Concatenate all POVM effects to treat as states for linear inversion
    states = []
    for index in range(size):
        for outcome in range(num_outcomes):
            states.append(basis.matrix((index,), outcome, (qubit,)))

    dual_basis = _construct_dual_states(states)

    # Organize back into nested lists of dual POVM effects
    dual_povms = []
    idx = 0
    for _ in range(size):
        dual_povms.append([dual_basis[idx + i] for i in range(num_outcomes)])
        idx += num_outcomes
    return dual_povms


def _construct_dual_states(states: Sequence[np.ndarray]):
    """Construct a dual preparation basis for linear inversion"""
    mats = np.asarray(states)
    size, dim1, dim2 = np.shape(mats)
    vec_basis = np.reshape(mats, (size, dim1 * dim2))
    basis_mat = np.sum([np.outer(i, np.conj(i)) for i in vec_basis], axis=0)

    try:
        inv_mat = np.linalg.inv(basis_mat)
    except np.linalg.LinAlgError as ex:
        raise ValueError(
            "Cannot construct dual basis states. Input states are not tomographically complete"
        ) from ex

    vec_dual = np.tensordot(inv_mat, vec_basis, axes=([1], [1])).T
    dual_mats = np.reshape(vec_dual, (size, dim1, dim2)).round(15)
    return dual_mats



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Fine amplitude characterization experiment."""

from typing import List, Optional, Sequence
import numpy as np

from qiskit import QuantumCircuit
from qiskit.circuit import Gate
from qiskit.circuit.library import XGate, SXGate
from qiskit.providers.backend import Backend
from qiskit_experiments.data_processing import DataProcessor, nodes
from qiskit_experiments.framework import BaseExperiment, Options
from qiskit_experiments.framework.restless_mixin import RestlessMixin
from qiskit_experiments.library.characterization.analysis import FineAmplitudeAnalysis


class FineAmplitude(BaseExperiment, RestlessMixin):
    r"""An experiment to determine the optimal pulse amplitude by amplifying gate errors.

    # section: overview

        The :class:`FineAmplitude` experiment repeats N times a gate with a pulse
        to amplify the under-/over-rotations in the gate to determine the optimal amplitude.
        The circuits are therefore of the form:

        .. parsed-literal::

                       ┌──────┐       ┌──────┐ ░ ┌─┐
                  q_0: ┤ Gate ├─ ... ─┤ Gate ├─░─┤M├
                       └──────┘       └──────┘ ░ └╥┘
            measure: 1/═════════ ... ═════════════╩═
                                                  0

        Here, Gate is the name of the gate which will be repeated. The user can optionally add a
        square-root of X pulse before the gates are repeated. This square-root of X pulse allows
        the analysis to differentiate between over rotations and under rotations in the case of
        :math:`\pi`-pulses. Importantly, the resulting data is analyzed by a fit to a cosine function in
        which we try to determine the over/under rotation given an intended rotation angle per
        gate which must also be specified by the user.

        Error amplifying experiments are most sensitive to angle errors when we measure points along
        the equator of the Bloch sphere. This is why users should insert a square-root of X pulse
        before running calibrations for :math:`\pm\pi` rotations. When all data points are close to
        the equator, it is difficult for a fitter to infer the overall scale of the error. When
        calibrating a :math:`\pi` rotation, one can use ``add_xp_circuit = True`` to insert one
        circuit that puts the qubit in the excited state to set the scale for the other circuits.
        Furthermore, when running calibrations for :math:`\pm\pi/2` rotations users are advised
        to use an odd number of repetitions, e.g. [1, 2, 3, 5, 7, ...] to ensure that the ideal
        points are on the equator of the Bloch sphere. Note the presence of two repetitions which
        allows us to prepare the excited state. Therefore, ``add_xp_circuit = True`` is not needed
        in this case.

    # section: example
        .. jupyter-execute::
            :hide-code:

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, 1e6, noise=True, seed=185)

        .. jupyter-execute::

            import numpy as np
            from qiskit.circuit.library import XGate
            from qiskit_experiments.library import FineAmplitude

            exp = FineAmplitude(physical_qubits=(0,), gate=XGate(), backend=backend)
            exp.analysis.set_options(fixed_parameters={"angle_per_gate" : np.pi, "phase_offset" : np.pi})

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: analysis_ref
        :class:`FineAmplitudeAnalysis`

    # section: reference
        .. ref_arxiv:: 1 1504.06597

    # section: manual
        :ref:`fine-amplitude-cal`

    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        r"""Default values for the fine amplitude experiment.

        Experiment Options:
            repetitions (List[int]): A list of the number of times that the gate is repeated.
            gate (Gate): This is a gate class such as XGate, so that one can obtain a gate
                by doing :code:`options.gate()`.
            normalization (bool): If set to True the DataProcessor will normalized the
                measured signal to the interval [0, 1]. Defaults to True.
            add_cal_circuits (bool): If set to True then two circuits to calibrate 0 and 1 points
                will be added. These circuits are often needed to properly calibrate the amplitude
                of the ping-pong oscillation that encodes the errors. This helps account for
                state preparation and measurement errors.
        """
        options = super()._default_experiment_options()
        options.repetitions = list(range(1, 15))
        options.gate = None
        options.normalization = True
        options.add_cal_circuits = True

        return options

    def __init__(
        self,
        physical_qubits: Sequence[int],
        gate: Gate,
        backend: Optional[Backend] = None,
        measurement_qubits: Sequence[int] = None,
    ):
        """Setup a fine amplitude experiment on the given qubit.

        Args:
            physical_qubits: Sequence containing qubit(s) on which to run the
                fine amplitude calibration experiment.
            gate: The gate that will be repeated.
            backend: Optional, the backend to run the experiment on.
            measurement_qubits: The qubits in the given physical qubits that need to
                be measured.
        """
        super().__init__(physical_qubits, analysis=FineAmplitudeAnalysis(), backend=backend)
        self.set_experiment_options(gate=gate)

        if measurement_qubits is not None:
            self._measurement_qubits = [self.physical_qubits.index(q) for q in measurement_qubits]
        else:
            self._measurement_qubits = range(self.num_qubits)

    def _spam_cal_circuits(self, meas_circuit: QuantumCircuit) -> List[QuantumCircuit]:
        """This method returns the calibration circuits.

        Calibration circuits allow the experiment to overcome state preparation and
        measurement errors which cause ideal probabilities to be below 1.

        Args:
            meas_circuit: The measurement circuit, so that we only apply x gates to the
                measured qubits.

        Returns:
            Two circuits that calibrate the spam errors for the 0 and 1 state.
        """
        cal_circuits = []

        for add_x in [0, 1]:
            circ = QuantumCircuit(self.num_qubits, meas_circuit.num_clbits)

            if add_x:
                qubits = meas_circuit.get_instructions("measure")[0].qubits
                circ.x(qubits)

            circ.compose(meas_circuit, inplace=True)

            circ.metadata = {
                "xval": add_x,
                "series": "spam-cal",
            }

            cal_circuits.append(circ)

        return cal_circuits

    def _pre_circuit(self, num_clbits: int) -> QuantumCircuit:
        """Return a preparation circuit.

        This method can be overridden by subclasses e.g. to calibrate gates on
        transitions other than the 0 <-> 1 transition.
        """
        return QuantumCircuit(self.num_qubits, num_clbits)

    def _measure_circuit(self) -> QuantumCircuit:
        """Create the measurement part of the quantum circuit.

        Sub-classes may override this function.

        Returns:
            A quantum circuit which defines the qubits that will be measured.
        """
        circuit = QuantumCircuit(self.num_qubits, len(self._measurement_qubits))

        for idx, qubit in enumerate(self._measurement_qubits):
            circuit.measure(qubit, idx)

        return circuit

    def circuits(self) -> List[QuantumCircuit]:
        """Create the circuits for the fine amplitude calibration experiment.

        Returns:
            A list of circuits with a variable number of gates.

        Raises:
            CalibrationError: If the analysis options do not contain the angle_per_gate.
        """
        repetitions = self.experiment_options.get("repetitions")

        qubits = range(self.num_qubits)
        meas_circ = self._measure_circuit()
        pre_circ = self._pre_circuit(meas_circ.num_clbits)

        if self.experiment_options.add_cal_circuits:
            circuits = self._spam_cal_circuits(meas_circ)
        else:
            circuits = []

        for repetition in repetitions:
            circuit = QuantumCircuit(self.num_qubits, meas_circ.num_clbits)

            # Add pre-circuit
            circuit.compose(pre_circ, qubits, range(meas_circ.num_clbits), inplace=True)

            for _ in range(repetition):
                circuit.append(self.experiment_options.gate, qubits)

            # Add the measurement part of the circuit
            circuit.compose(meas_circ, qubits, range(meas_circ.num_clbits), inplace=True)

            circuit.metadata = {
                "xval": repetition,
                "series": 1,
            }

            circuits.append(circuit)

        return circuits

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata


class FineXAmplitude(FineAmplitude):
    r"""A fine amplitude experiment with all the options set for the :math:`\pi`-rotation.

    # section: overview

        :class:`FineXAmplitude` is a subclass of :class:`FineAmplitude` and is used to set
        the appropriate values for the default options.

    # section: example
        .. jupyter-execute::
            :hide-code:

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, 1e4, noise=True, seed=198)

        .. jupyter-execute::

            from qiskit_experiments.library import FineXAmplitude

            exp = FineXAmplitude(physical_qubits=(0,), backend=backend)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    def __init__(self, physical_qubits: Sequence[int], backend: Optional[Backend] = None):
        """Initialize the experiment."""
        super().__init__(physical_qubits, XGate(), backend=backend)
        # Set default analysis options
        self.analysis.set_options(
            fixed_parameters={
                "angle_per_gate": np.pi,
                "phase_offset": np.pi / 2,
            }
        )

    @classmethod
    def _default_experiment_options(cls) -> Options:
        r"""Default values for the fine amplitude experiment.

        Experiment Options:
            gate (Gate): Gate to characterize. Defaults to an XGate.
        """
        options = super()._default_experiment_options()
        options.gate = XGate()
        return options

    def _pre_circuit(self, num_clbits: int) -> QuantumCircuit:
        """The preparation circuit is an sx gate to move to the equator of the Bloch sphere."""
        circuit = QuantumCircuit(self.num_qubits, num_clbits)
        circuit.sx(0)
        return circuit


class FineSXAmplitude(FineAmplitude):
    r"""A fine amplitude experiment with all the options set for the :math:`\pi/2`-rotation.

    # section: overview

        :class:`FineSXAmplitude` is a subclass of :class:`FineAmplitude` and is used to set
        the appropriate values for the default options.

    # section: example
        .. jupyter-execute::
            :hide-code:

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, 1e4, noise=True, seed=198)

        .. jupyter-execute::

            from qiskit_experiments.library import FineSXAmplitude

            exp = FineSXAmplitude(physical_qubits=(0,), backend=backend)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    def __init__(self, physical_qubits: Sequence[int], backend: Optional[Backend] = None):
        """Initialize the experiment."""
        super().__init__(physical_qubits, SXGate(), backend=backend)
        # Set default analysis options
        self.analysis.set_options(
            fixed_parameters={
                "angle_per_gate": np.pi / 2,
                "phase_offset": np.pi,
            }
        )

    @classmethod
    def _default_experiment_options(cls) -> Options:
        r"""Default values for the fine amplitude experiment.

        Experiment Options:
            gate (Gate): FineSXAmplitude calibrates an SXGate.
            add_cal_circuits (bool): If set to True then two circuits to calibrate 0 and 1 points
                will be added. This option is set to False by default for ``FineSXAmplitude``
                since the amplitude calibration can be achieved with two SX gates and this is
                included in the repetitions.
            repetitions (List[int]): By default the repetitions take on odd numbers for
                :math:`\pi/2` target angles as this ideally prepares states on the equator of
                the Bloch sphere. Note that the repetitions include two repetitions which
                plays the same role as including a circuit with an X gate.
        """
        options = super()._default_experiment_options()
        options.gate = SXGate()
        options.add_cal_circuits = False
        options.repetitions = [0, 1, 2, 3, 5, 7, 9, 11, 13, 15, 17, 21, 23, 25]

        return options


class FineZXAmplitude(FineAmplitude):
    r"""A fine amplitude experiment for the :code:`RZXGate(np.pi / 2)`.

    # section: overview

        :class:`FineZXAmplitude` is a subclass of :class:`FineAmplitude` and is used to set
        the appropriate values for the default options to calibrate a :code:`RZXGate(np.pi / 2)`.

    # section: example

        To run this experiment, the user will have to provide the instruction schedule
        map in the transpile options that contains the schedule for the experiment.

        .. code-block:: python

            qubits = (1, 2)
            inst_map = InstructionScheduleMap()
            inst_map.add("szx", qubits, my_schedule)

            fine_amp = FineZXAmplitude(qubits, backend)
            fine_amp.set_transpile_options(inst_map=inst_map)

        Here, :code:`my_schedule` is the pulse schedule that will implement the
        :code:`RZXGate(np.pi / 2)` rotation.
    """

    def __init__(self, physical_qubits: Sequence[int], backend: Optional[Backend] = None):
        """Initialize the experiment."""

        # We cannot use RZXGate since it has a parameter so we redefine the gate.
        # Failing to do so causes issues with QuantumCircuit.calibrations.
        gate = Gate("szx", 2, [])

        super().__init__(
            physical_qubits, gate, backend=backend, measurement_qubits=[physical_qubits[1]]
        )
        # Set default analysis options
        self.analysis.set_options(
            fixed_parameters={
                "angle_per_gate": np.pi / 2,
                "phase_offset": np.pi,
            },
            outcome="1",
        )

    @classmethod
    def _default_experiment_options(cls) -> Options:
        r"""Default values for the fine amplitude experiment.

        Experiment Options:
            add_cal_circuits (bool): If set to True then two circuits to calibrate 0 and 1 points
                will be added. This option is set to False by default for ``FineZXAmplitude``
                since the amplitude calibration can be achieved with two RZX gates and this is
                included in the repetitions.
            repetitions (List[int]): A list of the number of times that the gate is repeated.
        """
        options = super()._default_experiment_options()
        options.add_cal_circuits = False
        options.repetitions = [0, 1, 2, 3, 4, 5, 7, 9, 11, 13]
        return options

    @classmethod
    def _default_transpile_options(cls) -> Options:
        """Default transpile options for the fine amplitude experiment.

        Experiment Options:
            basis_gates: Set to :code:`["szx"]`.
            inst_map: The instruction schedule map that will contain the schedule for the
                Rzx(pi/2) gate. This schedule should be stored under the instruction name
                ``szx``.
        """
        options = super()._default_transpile_options()
        options.basis_gates = ["szx"]
        options.inst_map = None
        return options

    def enable_restless(
        self,
        rep_delay: Optional[float] = None,
        override_processor_by_restless: bool = True,
        suppress_t1_error: bool = False,
    ):
        """Enable restless measurements.

        We wrap the method of the :class:`.RestlessMixin` to readout both qubits. This forces
        the control qubit to be in either the 0 or 1 state before the next circuit starts
        since restless measurements do not reset qubits.

        Args:
            rep_delay: The repetition delay. This is the delay between a measurement
                and the subsequent quantum circuit. Since the backends have
                dynamic repetition rates, the repetition delay can be set to a small
                value which is required for restless experiments. Typical values are
                1 us or less.
            override_processor_by_restless: If False, a data processor that is specified in the
                analysis options of the experiment is not overridden by the restless data
                processor. The default is True.
            suppress_t1_error: If True, the default is False, then no error will be raised when
                ``rep_delay`` is larger than the T1 times of the qubits. Instead, a warning will
                be logged as restless measurements may have a large amount of noise.
        """
        self.analysis.set_options(outcome="11")
        super().enable_restless(rep_delay, override_processor_by_restless, suppress_t1_error)
        self._measurement_qubits = range(self.num_qubits)

    def _get_restless_processor(self, meas_level: int = 2) -> DataProcessor:
        """Marginalize the counts after the restless shot reordering."""
        return DataProcessor(
            "memory",
            [
                nodes.RestlessToCounts(self._num_qubits),
                nodes.MarginalizeCounts({1}),  # keep only the target.
                nodes.Probability("1"),
            ],
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Quantum Volume Experiment class.
"""

from typing import Union, Sequence, Optional, List
from numpy.random import Generator, default_rng
from numpy.random.bit_generator import BitGenerator, SeedSequence

from qiskit.utils.optionals import HAS_AER

from qiskit import QuantumCircuit
from qiskit.circuit.library import QuantumVolume as QuantumVolumeCircuit
from qiskit import transpile
from qiskit.providers.backend import Backend
from qiskit_experiments.framework import BaseExperiment, Options
from .qv_analysis import QuantumVolumeAnalysis


class QuantumVolume(BaseExperiment):
    """An experiment to measure the largest random square circuit that can be run on a processor.

    # section: overview
        Quantum Volume (QV) is a single-number metric that can be measured using a concrete protocol
        on near-term quantum computers of modest size. The QV method quantifies the largest random
        circuit of equal width and depth that the computer successfully implements.
        Quantum computing systems with high-fidelity operations, high connectivity,
        large calibrated gate sets, and circuit rewriting toolchains are expected to
        have higher quantum volumes.

        The Quantum Volume is determined by the largest circuit depth :math:`d_{max}`,
        and equals to :math:`2^{d_{max}}`.
        See the `Qiskit Textbook
        <https://github.com/Qiskit/textbook/blob/main/notebooks/quantum-hardware/measuring-quantum-volume.ipynb>`_
        for an explanation on the QV protocol.

        In the QV experiment we generate :class:`~qiskit.circuit.library.QuantumVolume` circuits on
        :math:`d` qubits, which contain :math:`d` layers, where each layer consists of random 2-qubit
        unitary gates from :math:`SU(4)`, followed by a random permutation on the :math:`d` qubits.
        Then these circuits run on the quantum backend and on an ideal simulator (either
        :class:`~qiskit_aer.AerSimulator` or :class:`~qiskit.quantum_info.Statevector`).

        A depth :math:`d` QV circuit is successful if it has `mean heavy-output probability` > 2/3 with
        confidence level > 0.977 (corresponding to z_value = 2), and at least 100 trials have been ran.

        See :class:`QuantumVolumeAnalysis` documentation for additional
        information on QV experiment analysis.

    # section: analysis_ref
        :class:`QuantumVolumeAnalysis`

    # section: manual
        :doc:`/manuals/verification/quantum_volume`

    # section: reference
        .. ref_arxiv:: 1 1811.12926
        .. ref_arxiv:: 2 2008.08571

    """

    def __init__(
        self,
        physical_qubits: Sequence[int],
        backend: Optional[Backend] = None,
        trials: Optional[int] = 100,
        seed: Optional[Union[int, SeedSequence, BitGenerator, Generator]] = None,
        simulation_backend: Optional[Backend] = None,
    ):
        """Initialize a quantum volume experiment.

        Args:
            physical_qubits: list of physical qubits for the experiment.
            backend: Optional, the backend to run the experiment on.
            trials: The number of trials to run the quantum volume circuit.
            seed: Optional, seed used to initialize ``numpy.random.default_rng``
                  when generating circuits. The ``default_rng`` will be initialized
                  with this seed value every time :meth:`circuits` is called.
            simulation_backend: The simulator backend to use to generate
                the expected results. the simulator must have a 'save_probabilities'
                method. If None, the :class:`qiskit_aer.AerSimulator` simulator will be used
                (in case :external+qiskit_aer:doc:`qiskit-aer <index>` is not
                installed, :class:`qiskit.quantum_info.Statevector` will be used).
        """
        super().__init__(physical_qubits, analysis=QuantumVolumeAnalysis(), backend=backend)

        # Set configurable options
        self.set_experiment_options(trials=trials, seed=seed)

        if not simulation_backend and HAS_AER:
            from qiskit_aer import AerSimulator

            self._simulation_backend = AerSimulator()
        else:
            self._simulation_backend = simulation_backend

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            trials (int): Optional, number of times to generate new Quantum Volume
                circuits and calculate their heavy output.
            seed (None or int or SeedSequence or BitGenerator or Generator): A seed
                used to initialize ``numpy.random.default_rng`` when generating circuits.
                The ``default_rng`` will be initialized with this seed value every time
                :meth:`circuits` is called.
        """
        options = super()._default_experiment_options()

        options.trials = 100
        options.seed = None

        return options

    def _get_ideal_data(self, circuit: QuantumCircuit, **run_options) -> List[float]:
        """Return ideal measurement probabilities.

        In case the user does not have Aer installed, use Qiskit's quantum info module
        to calculate the ideal state.

        Args:
            circuit: the circuit to extract the ideal data from
            run_options: backend run options.

        Returns:
            list: list of the probabilities for each state in the circuit.
        """
        ideal_circuit = circuit.remove_final_measurements(inplace=False)
        if self._simulation_backend:
            ideal_circuit.save_probabilities()
            # always transpile with optimization_level 0, even if the non ideal circuits will run
            # with different optimization level, because we need to compare the results to the
            # exact generated probabilities
            ideal_circuit = transpile(ideal_circuit, self._simulation_backend, optimization_level=0)

            ideal_result = self._simulation_backend.run(ideal_circuit, **run_options).result()
            probabilities = ideal_result.data().get("probabilities")
        else:
            from qiskit.quantum_info import Statevector

            state_vector = Statevector(ideal_circuit)
            probabilities = state_vector.probabilities()
        return list(probabilities)

    def circuits(self) -> List[QuantumCircuit]:
        """Return a list of Quantum Volume circuits.

        Returns:
            A list of :class:`QuantumCircuit`.
        """
        rng = default_rng(seed=self.experiment_options.seed)
        circuits = []
        depth = self._num_qubits

        # Note: the trials numbering in the metadata is starting from 1 for each new experiment run
        for trial in range(1, self.experiment_options.trials + 1):
            qv_circ = QuantumVolumeCircuit(depth, depth, seed=rng)
            qv_circ.measure_active()
            qv_circ.metadata = {
                "depth": depth,
                "trial": trial,
                "ideal_probabilities": self._get_ideal_data(qv_circ),
            }
            circuits.append(qv_circ)
        return circuits



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Pauli preparation and measurement tomography bases.
"""
from typing import Optional
import numpy as np
from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library import HGate, XGate, ZGate, SGate, SdgGate
from qiskit.result import LocalReadoutMitigator
from qiskit.quantum_info import DensityMatrix
from qiskit.exceptions import QiskitError
from .local_basis import LocalMeasurementBasis, LocalPreparationBasis


class PauliMeasurementBasis(LocalMeasurementBasis):
    r"""Standard Pauli measurement basis.

    This basis has 3 indices each with with 2 measurement outcomes. The
    corresponding single-qubit measurement circuits and outcome POVM
    matrices are:

    .. table:: Single-qubit measurement circuits and POVM matrices

        +-------+-------+---------------+---------+-------------------------------+
        | Index | Basis | Circuit       | Outcome | POVM Matrix                   |
        +=======+=======+===============+=========+===============================+
        | 0     | Z     |``-[I]-``      | 0       |``[[1, 0], [0, 0]]``           |
        +-------+-------+---------------+---------+-------------------------------+
        |       |       |               | 1       |``[[0, 0], [0, 1]]``           |
        +-------+-------+---------------+---------+-------------------------------+
        | 1     | X     | ``-[H]-``     | 0       |``[[0.5, 0.5], [0.5, 0.5]]``   |
        +-------+-------+---------------+---------+-------------------------------+
        |       |       |               | 1       |``[[0.5, -0.5], [-0.5, 0.5]]`` |
        +-------+-------+---------------+---------+-------------------------------+
        | 2     | Y     |``-[SDG]-[H]-``| 0       |``[[0.5, -0.5j], [0.5j, 0.5]]``|
        +-------+-------+---------------+---------+-------------------------------+
        |       |       |               | 1       |``[[0.5, 0.5j], [-0.5j, 0.5]]``|
        +-------+-------+---------------+---------+-------------------------------+

    """

    def __init__(self, mitigator: Optional[LocalReadoutMitigator] = None):
        """Initialize Pauli measurement basis"""
        # Z-meas rotation
        meas_z = QuantumCircuit(1, name="PauliMeasZ")
        # X-meas rotation
        meas_x = QuantumCircuit(1, name="PauliMeasX")
        meas_x.append(HGate(), [0])
        # Y-meas rotation
        meas_y = QuantumCircuit(1, name="PauliMeasY")
        meas_y.append(SdgGate(), [0])
        meas_y.append(HGate(), [0])
        instructions = [meas_z, meas_x, meas_y]
        self._mitigator = mitigator

        super().__init__(
            "PauliMeasurementBasis", instructions, qubit_povms=self._mitigator_povm(self._mitigator)
        )

    @staticmethod
    def _mitigator_povm(mitigator: Optional[LocalReadoutMitigator] = None):
        """Construct LocalMeasurementBasis qubit_povm from mitigator."""
        if mitigator is None:
            return None
        povm = {}
        if isinstance(mitigator, LocalReadoutMitigator):
            povm = {}
            for qubit in mitigator.qubits:
                amat = mitigator.assignment_matrix(qubit)
                povm[(qubit,)] = {(0,): [DensityMatrix(np.diag(row)) for row in amat]}
        else:
            raise QiskitError("Invalid mitigator: must be LocalReadoutMitigator")
        return povm

    def __json_encode__(self):
        # Override LocalMeasurementBasis's encoder
        if self._mitigator is not None:
            return {"mitigator": self._mitigator}
        return {}


class PauliPreparationBasis(LocalPreparationBasis):
    """Minimal 4-element Pauli measurement basis.

    This is a minimal size 4 preparation basis where each qubit
    index corresponds to the following initial state preparation
    circuits and density matrices:

    .. table:: Single-qubit preparation circuits and states

        +-------+-------+---------------------+---------------------------------+
        | Index | State | Preparation Circuit | Density Matrix                  |
        +=======+=======+=====================+=================================+
        | 0     | Zp    | ``-[I]-``           | ``[[1, 0], [0, 0]]``            |
        +-------+-------+---------------------+---------------------------------+
        | 1     | Zm    | ``-[X]-``           | ``[[0, 0], [0, 1]]``            |
        +-------+-------+---------------------+---------------------------------+
        | 2     | Xp    | ``-[H]-``           | ``[[0.5, 0.5], [0.5, 0.5]]``    |
        +-------+-------+---------------------+---------------------------------+
        | 3     | Yp    | ``-[H]-[S]-``       | ``[[0.5, -0.5j], [0.5j, 0.5]]`` |
        +-------+-------+---------------------+---------------------------------+

    """

    def __init__(self):
        """Initialize Pauli preparation basis"""
        # |0> Zp rotation
        prep_zp = QuantumCircuit(1, name="PauliPrepZp")
        # |1> Zm rotation
        prep_zm = QuantumCircuit(1, name="PauliPrepZm")
        prep_zm.append(XGate(), [0])
        # |+> Xp rotation
        prep_xp = QuantumCircuit(1, name="PauliPrepXp")
        prep_xp.append(HGate(), [0])
        # |+i> Yp rotation
        prep_yp = QuantumCircuit(1, name="PauliPrepYp")
        prep_yp.append(HGate(), [0])
        prep_yp.append(SGate(), [0])
        super().__init__("PauliPreparationBasis", [prep_zp, prep_zm, prep_xp, prep_yp])

    def __json_encode__(self):
        # Override LocalPreparationBasis's encoder
        return {}


class Pauli6PreparationBasis(LocalPreparationBasis):
    """Over-complete 6-element Pauli preparation basis.

    This is an over-complete size 6 preparation basis where each qubit
    index corresponds to the following initial state density matrices:

    .. table:: Single-qubit preparation circuits and states

        +-------+-------+---------------------+---------------------------------+
        | Index | State | Preparation Circuit | Density Matrix                  |
        +=======+=======+=====================+=================================+
        | 0     | Zp    | ``-[I]-``           | ``[[1, 0], [0, 0]]``            |
        +-------+-------+---------------------+---------------------------------+
        | 1     | Zm    | ``-[X]-``           | ``[[0, 0], [0, 1]]``            |
        +-------+-------+---------------------+---------------------------------+
        | 2     | Xp    | ``-[H]-``           | ``[[0.5, 0.5], [0.5, 0.5]]``    |
        +-------+-------+---------------------+---------------------------------+
        | 3     | Xm    | ``-[H]-[Z]-``       | ``[[0.5, -0.5], [-0.5, 0.5]]``  |
        +-------+-------+---------------------+---------------------------------+
        | 2     | Yp    | ``-[H]-[S]-``       | ``[[0.5, -0.5j], [0.5j, 0.5]]`` |
        +-------+-------+---------------------+---------------------------------+
        | 3     | Ym    | ``-[H]-[Sdg]-``     | ``[[0.5, 0.5j], [-0.5j, 0.5]]`` |
        +-------+-------+---------------------+---------------------------------+

    """

    def __init__(self):
        # |0> Zp rotation
        prep_zp = QuantumCircuit(1, name="PauliPrepZp")
        # |1> Zm rotation
        prep_zm = QuantumCircuit(1, name="PauliPrepZm")
        prep_zm.append(XGate(), [0])
        # |+> Xp rotation
        prep_xp = QuantumCircuit(1, name="PauliPrepXp")
        prep_xp.append(HGate(), [0])
        # |-> Xm rotation
        prep_xm = QuantumCircuit(1, name="PauliPrepXm")
        prep_xm.append(HGate(), [0])
        prep_xm.append(ZGate(), [0])
        # |+i> Yp rotation
        prep_yp = QuantumCircuit(1, name="PauliPrepYp")
        prep_yp.append(HGate(), [0])
        prep_yp.append(SGate(), [0])
        # |-i> Ym rotation
        prep_ym = QuantumCircuit(1, name="PauliPrepYm")
        prep_ym.append(HGate(), [0])
        prep_ym.append(SdgGate(), [0])
        super().__init__(
            "Pauli6PreparationBasis",
            [
                prep_zp,
                prep_zm,
                prep_xp,
                prep_xm,
                prep_yp,
                prep_ym,
            ],
        )

    def __json_encode__(self):
        # Override LocalPreparationBasis's encoder
        return {}



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Plotter for IQ data."""
import warnings
from itertools import product
from typing import List, Optional, Tuple

import numpy as np

from qiskit_experiments.data_processing.discriminator import BaseDiscriminator
from qiskit_experiments.framework import Options

from ..utils import DataExtentCalculator, ExtentTuple
from .base_plotter import BasePlotter


class IQPlotter(BasePlotter):
    """A plotter class to plot IQ data.

    :class:`.IQPlotter` plots results from experiments which used measurement-level 1,
    i.e. IQ data. This class also supports plotting predictions from a discriminator
    (subclass of :class:`.BaseDiscriminator`), which is used to classify IQ results into
    labels. The discriminator labels are matched with the series names to generate an
    image of the predictions. Points that are misclassified by the discriminator are
    flagged in the figure (see the ``flag_misclassified`` option). A canonical
    application of :class:`.IQPlotter` is for classification of single-qubit readout for
    different prepared states.

    Example:
        .. code-block:: python

            # Create plotter
            plotter = IQPlotter(MplDrawer())

            # Iterate over results, one per prepared state. Add points and centroid to
            # plotter and set label for prepared states as |n> where n is the
            # prepared-state number.
            series_params = {}
            for res in results:
                # Get IQ points from result memory.
                points = res.memory

                # Compute centroid as mean of all points.
                centroid = np.mean(points, axis=1)

                # Get ``prep``, which is part of the result metadata.
                prep = res.prep

                # Create label as a ket instead of just a state number (i.e., prep).
                series_params[prep] = {
                    "label":f"|{prep}>",
                }

                plotter.set_series_data(prep, points=points, centroid=centroid)
            plotter.set_figure_options(series_params=series_params)
            ...
            # Optional: Add trained discriminator.
            discrim = MyIQDiscriminator()
            # Discriminator labels are the same as series names.
            discrim.fit(train_data, train_labels)
            plotter.set_supplementary_data(discriminator=discrim)
            ...
            # Plot figure.
            fig = plotter.figure()
    """

    @classmethod
    def expected_series_data_keys(cls) -> List[str]:
        """Returns the expected series data keys supported by this plotter.

        Data Keys:
            points: Single-shot IQ data.
            centroid: Averaged IQ data.
        """
        return [
            "points",
            "centroid",
        ]

    @classmethod
    def expected_supplementary_data_keys(cls) -> List[str]:
        """Returns the expected figures data keys supported by this plotter.

        Data Keys:
            discriminator: A trained discriminator that classifies IQ points. If
                provided, the predictions of the discriminator will be sampled to
                generate a background image, indicating the regions for each predicted
                outcome. The predictions are assumed to be series names (``Union[str,
                int, float]``). The generated image allows viewers to see how well the
                discriminator classifies the provided series data. Must be a subclass of
                :class:`.BaseDiscriminator`. See :attr:`options` for ways to control the
                generation of the discriminator prediction image.
            fidelity: A float representing the fidelity of the discrimination.
        """
        return [
            "discriminator",
            "fidelity",
        ]

    def _compute_extent(self) -> Optional[ExtentTuple]:
        """Computes the extent tuple of the data being plotted.

        Returns:
            The tuple ``(x_min, x_max, y_min, y_max)``, defining a rectangle containing
            all the data for this plotter. If the plotter contains no data, ``None`` is
            returned instead.
        """
        ext_calc = DataExtentCalculator(
            multiplier=self.options.discriminator_multiplier,
            aspect_ratio=self.options.discriminator_aspect_ratio,
        )
        has_registered_data = False
        for series in self.series:
            if self.data_exists_for(series, "points"):
                (points,) = self.data_for(series, "points")
                ext_calc.register_data(points)
                has_registered_data = True
            if self.data_exists_for(series, "centroid"):
                (centroid,) = self.data_for(series, "centroid")
                ext_calc.register_data(np.asarray(centroid).reshape(1, 2))
                has_registered_data = True
        if self.figure_options.xlim:
            ext_calc.register_data(self.figure_options.xlim, dim=0)
            has_registered_data = True
        if self.figure_options.ylim:
            ext_calc.register_data(self.figure_options.ylim, dim=1)
            has_registered_data = True
        if has_registered_data:
            return ext_calc.extent()
        else:
            return None

    def _compute_discriminator_image(
        self,
    ) -> Tuple[Optional[np.ndarray], Optional[ExtentTuple]]:
        """Compute the array/image sampled from the discriminator predictions.

        Returns:
            The tuple ``(img, extent)`` where ``img`` is an optional 2D NumPy array of
            predictions and ``extent`` is a tuple of the extent ``(x_min, x_max, y_min,
            y_max)`` of the image.
        """
        # If the discriminator is not provided, cannot compute the image.
        if "discriminator" not in self.supplementary_data:
            return None, None

        if self.options.discriminator_extent:
            extent = self.options.discriminator_extent
        else:
            extent = self._compute_extent()
        # If ``extent`` is None, we don't have data to plot and thus cannot generate the
        # discriminator image.
        if extent is None:
            return None, None

        # Get the discriminator and check if it is trained. If not, return.
        discrim: BaseDiscriminator = self.supplementary_data["discriminator"]
        if not discrim.is_trained():
            return None, None

        # Compute discriminator resolution.
        extent_range = np.diff(np.asarray(extent).reshape(2, 2), axis=1).flatten()
        resolution = (
            (extent_range / np.max(extent_range)) * self.options.discriminator_max_resolution
        ).astype(int)

        # Create coordinates for each pixel in the image, in the same units as `extent`.
        coords = [
            [x, y]
            for x, y in product(
                np.linspace(extent[0], extent[1], resolution[0]),
                np.linspace(extent[2], extent[3], resolution[1]),
            )
        ]

        # Get predictions for coordinates from the discriminator.
        predictions = discrim.predict(coords)

        # Unwrap predictions into a 2D array
        predictions = np.reshape(predictions, tuple(resolution))

        return predictions, extent

    @classmethod
    def _default_options(cls) -> Options:
        """Return iq-plotter specific default plotter options.

        Options:
            plot_discriminator (bool): Whether to plot an image showing the predictions
                of the ``discriminator`` entry in :attr:`supplementary_data`. If True,
                the "discriminator" supplementary data entry must be set.
            discriminator_multiplier (float): The multiplier to use when computing the
                extent of the discriminator plot. The range of the series data is taken
                as the base value and multiplied by ``discriminator_extent_multiplier``
                to compute the extent of the discriminator predictions. Defaults to 1.1.
            discriminator_aspect_ratio (float): The aspect ratio of the extent of the
                discriminator predictions, being ``width/height``. Defaults to ``1`` for
                a square region.
            discriminator_max_resolution (int): The number of pixels to use for the
                largest edge of the discriminator extent, used when sampling the
                discriminator to create the prediction image. Defaults to 1024.
            discriminator_alpha (float): The transparency of the discriminator
                prediction image. Defaults to 0.2 (i.e., 20%).
            discriminator_extent (Optional[ExtentTuple]): An optional tuple defining the
                extent of the image created by sampling from the discriminator. If
                ``None``, the extent tuple is computed using
                ``discriminator_multiplier``, ``discriminator_aspect_ratio``, and the
                series-data ``points`` and ``centroid``. Defaults to ``None``.
            flag_misclassified (bool): Whether to mark misclassified IQ values from all
                ``points`` series data, based on whether their series name is not the
                same as the prediction from the discriminator provided as supplementary
                data. If ``discriminator`` is not provided, ``flag_misclassified`` has
                no effect. Defaults to True.
            misclassified_symbol (str): Symbol for misclassified points, as a
                drawer-compatible string. Defaults to "x".
            misclassified_color (str | tuple): Color for misclassified points, as a
                drawer-compatible string or RGB tuple. Defaults to "r".

        """
        options = super()._default_options()
        # Discriminator options
        options.plot_discriminator = True
        options.discriminator_multiplier = 1.1
        options.discriminator_aspect_ratio = 1.0
        options.discriminator_max_resolution = 1024
        options.discriminator_alpha = 0.2
        options.discriminator_extent = None
        # Points options
        options.flag_misclassified = True
        options.misclassified_symbol = "x"
        options.misclassified_color = "r"
        return options

    @classmethod
    def _default_figure_options(cls) -> Options:
        fig_opts = super()._default_figure_options()
        fig_opts.xlabel = "In-Phase"
        fig_opts.ylabel = "Quadrature"
        fig_opts.xval_unit = "arb."
        fig_opts.yval_unit = "arb."
        fig_opts.xval_unit_scale = False
        fig_opts.yval_unit_scale = False
        return fig_opts

    def _misclassified_points(self, series_name: str, points: np.ndarray) -> Optional[np.ndarray]:
        """Returns a list of IQ coordinates for points that are misclassified by the discriminator.

        Args:
            series_name: The series name to use as the expected discriminator label. If
                the discriminator returns a prediction that doesn't equal
                ``series_name``, it is marked as misclassified.
            points: The list of points to check for misclassification.

        Returns:
            A NumPy array of IQ points, being those that were misclassified by the
            discriminator. If the discriminator isn't set and trained, then `None` is
            returned. The array may be empty.
        """
        # Check if we have a discriminator, and if it is trained. If not, return None.
        if "discriminator" not in self.supplementary_data:
            return None
        discrim: BaseDiscriminator = self.supplementary_data["discriminator"]
        if not discrim.is_trained():
            return None
        classifications = discrim.predict(points)
        misclassified = np.argwhere(classifications != series_name)
        return points[misclassified, :].reshape(-1, 2)

    def _plot_figure(self):
        """Plots an IQ figure."""
        # Plot discriminator first so that subsequent graphics change the automatic limits. This is a
        # function of the way `imshow` works for Matplotlib, in that the limits are automatically changed
        # to the extents of the image being plotted. If `image` is called after `scatter`, then the
        # automatic axis limits will be set to match the extent of the image.
        if "discriminator" in self.supplementary_data and self.options.plot_discriminator:
            if self.options.subplots != (1, 1):
                warnings.warn(
                    "Plotting discriminator predictions with subplots is not well supported by "
                    "IQPlotter as the predictions image will only be drawn on one subplot."
                )
            discrim, extent = self._compute_discriminator_image()
            if discrim is None:
                warnings.warn(
                    "Discriminator was provided but the sampled predictions image could not be "
                    "generated."
                )
            else:
                self.drawer.image(
                    np.flip(discrim.transpose(), axis=0),
                    extent,
                    name="discriminator",
                    cmap_use_series_colors=True,
                    alpha=self.options.discriminator_alpha,
                )
        # Plot points and centroids
        for ser in self.series:
            has_plotted_centroid = False
            if self.data_exists_for(ser, "centroid"):
                (centroid,) = self.data_for(ser, "centroid")
                self.drawer.scatter(
                    centroid[0],
                    centroid[1],
                    name=ser,
                    legend=True,
                    zorder=4,
                    s=20,
                    edgecolor="k",
                    marker="o",
                )
                has_plotted_centroid = True
            if self.data_exists_for(ser, "points"):
                (points,) = self.data_for(ser, "points")
                self.drawer.scatter(
                    points[:, 0],
                    points[:, 1],
                    name=ser,
                    legend=not has_plotted_centroid,
                    zorder=2,
                    s=10,
                    alpha=0.2,
                    marker=".",
                )
                if self.options.flag_misclassified:
                    misclassified_points = self._misclassified_points(ser, points)
                    if misclassified_points is not None:
                        self.drawer.scatter(
                            misclassified_points[:, 0],
                            misclassified_points[:, 1],
                            name="misclassified",
                            legend=False,
                            zorder=3,
                            s=10,
                            alpha=0.4,
                            marker=self.options.misclassified_symbol,
                            color=self.options.misclassified_color,
                        )

        # Fidelity report
        report = self._write_report()
        if len(report) > 0:
            self.drawer.textbox(report)

    def _write_report(self) -> str:
        """Write fidelity report with supplementary_data.

        Subclass can override this method to customize fidelity report. By default, this
        writes the fidelity of the discriminator in the fidelity report.

        Returns:
            Fidelity report.
        """
        report = ""

        if "fidelity" in self.supplementary_data:
            fidelity = self.supplementary_data["fidelity"]
            report += f"fidelity = {fidelity: .4g}"

        return report



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Local readout error calibration experiment class.
"""
from typing import Iterable, List, Optional
from qiskit import QuantumCircuit
from qiskit.providers.backend import BackendV2, Backend
from qiskit.exceptions import QiskitError
from qiskit_experiments.framework import BaseExperiment
from qiskit_experiments.library.characterization.analysis.local_readout_error_analysis import (
    LocalReadoutErrorAnalysis,
)
from .correlated_readout_error import calibration_circuit


class LocalReadoutError(BaseExperiment):
    r"""An experiment for characterizing local readout error.

    # section: overview

        This class constructs a :class:`~qiskit.result.LocalReadoutMitigator` containing a sequence
        of assignment matrices :math:`A` characterizing the readout error for the given qubits
        from the experiment results. The full assignment matrix is accessible via the
        :meth:`~qiskit.result.LocalReadoutMitigator.assignment_matrix` method.

        Readout errors affect quantum computation during the measurement
        of the qubits in a quantum device. By characterizing the readout errors,
        it is possible to construct a *readout error mitigator* that is used both
        to obtain a more accurate distribution of the outputs, and more accurate
        measurements of expectation value for measurables.

        The readout mitigator is generated from an *assignment matrix*:
        a :math:`2^n \times 2^n` matrix :math:`A` such that :math:`A_{y,x}` is the probability
        to observe :math:`y` given the true outcome should be :math:`x`. The assignment matrix is used
        to compute the *mitigation matrix* used in the readout error mitigation process itself.

        A *Local readout mitigator* works under the assumption that readout errors are mostly
        *local*, meaning readout errors for different qubits are independent of each other.
        In this case, the assignment matrix is the tensor product of :math:`n` :math:`2 \times 2`
        matrices, one for each qubit, making it practical to store the assignment matrix in implicit
        form, by storing the individual :math:`2 \times 2` assignment matrices.
        The corresponding class in Qiskit is the :class:`~qiskit.result.LocalReadoutMitigator`
        in :mod:`qiskit.result`.

        The experiment generates 2 circuits, corresponding to the states
        :math:`|0^n\rangle` and :math:`|1^n\rangle`, measuring the error in all
        the qubits at once, and constructs the assignment matrix and local mitigator from the results.

        See :class:`LocalReadoutErrorAnalysis`
        documentation for additional information on local readout error experiment analysis.

    # section: analysis_ref
        :class:`LocalReadoutErrorAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
            from qiskit_experiments.test.patching import patch_sampler_test_support
            patch_sampler_test_support()

            # backend
            from qiskit_aer import AerSimulator
            from qiskit_ibm_runtime.fake_provider import FakePerth
            backend = AerSimulator.from_backend(FakePerth())

        .. jupyter-execute::

            from qiskit_experiments.library import LocalReadoutError

            qubits = list(range(4))
            exp = LocalReadoutError(physical_qubits=qubits, backend=backend)
            exp.analysis.set_options(plot=True)
            exp.set_run_options(shots=10000)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: manual
        :doc:`/manuals/measurement/readout_mitigation`

    # section: reference
        .. ref_arxiv:: 1 2006.14044
    """

    def __init__(
        self,
        physical_qubits: Optional[Iterable[int]] = None,
        backend: Optional[Backend] = None,
    ):
        """Initialize a local readout error characterization experiment.

        Args:
            physical_qubits: Optional, the backend qubits being characterized
                for readout error. If None all qubits on the provided backend
                will be characterized.
            backend: Optional, the backend to characterize.

        Raises:
            QiskitError: If args are not valid.
        """
        if physical_qubits is None:
            if backend is None:
                raise QiskitError("`physical_qubits` and `backend` kwargs cannot both be None.")
            num_qubits = 0
            if isinstance(backend, BackendV2):
                num_qubits = backend.target.num_qubits
            elif isinstance(backend, Backend):
                num_qubits = backend.configuration().num_qubits
            if num_qubits:
                physical_qubits = range(num_qubits)
            else:
                raise QiskitError(f"Cannot infer backend qubits from backend {backend}")
        super().__init__(physical_qubits, backend=backend)
        self.analysis = LocalReadoutErrorAnalysis()

    def circuits(self) -> List[QuantumCircuit]:
        """Returns the experiment's circuits"""
        labels = ["0" * self.num_qubits, "1" * self.num_qubits]
        return [calibration_circuit(self.num_qubits, label) for label in labels]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
T2Ramsey Experiment class.
"""
from typing import Union
import qiskit_experiments.curve_analysis as curve
from qiskit_experiments.framework import Options


class T2RamseyAnalysis(curve.DampedOscillationAnalysis):
    """T2 Ramsey result analysis class."""

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options."""
        options = super()._default_options()
        options.plotter.set_figure_options(
            xlabel="Delay",
            ylabel="P(1)",
            xval_unit="s",
        )

        options.result_parameters = [
            curve.ParameterRepr("freq", "Frequency", "Hz"),
            curve.ParameterRepr("tau", "T2star", "s"),
        ]

        return options

    def _evaluate_quality(self, fit_data: curve.CurveFitResult) -> Union[str, None]:
        """Algorithmic criteria for whether the fit is good or bad.

        A good fit has:
            - a reduced chi-squared lower than three and greater than zero
            - relative error of amp is less than 10 percent
            - relative error of tau is less than 10 percent
            - relative error of freq is less than 10 percent
        """
        amp = fit_data.ufloat_params["amp"]
        tau = fit_data.ufloat_params["tau"]
        freq = fit_data.ufloat_params["freq"]

        criteria = [
            0 < fit_data.reduced_chisq < 3,
            curve.utils.is_error_not_significant(amp, fraction=0.1),
            curve.utils.is_error_not_significant(tau, fraction=0.1),
            curve.utils.is_error_not_significant(freq, fraction=0.1),
        ]

        if all(criteria):
            return "good"

        return "bad"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
ZZ Ramsey experiment
"""

from typing import List, Tuple, Union

import numpy as np

from qiskit import QuantumCircuit
from qiskit.providers.backend import Backend
from qiskit.circuit import Parameter, ParameterExpression

from qiskit_experiments.framework import BackendTiming, BaseExperiment, Options
from .analysis.zz_ramsey_analysis import ZZRamseyAnalysis


class ZZRamsey(BaseExperiment):
    r"""An experiment to characterize the static :math:`ZZ` interaction for a qubit pair.

    # section: overview

        This experiment assumes a two qubit Hamiltonian of the form

        .. math::

            H = h \left(\frac{f_0}{2} ZI + \frac{f_1}{2} IZ + \frac{f_{ZZ}}{4} ZZ\right)

        and measures the strength :math:`f_{ZZ}` of the :math:`ZZ` term.
        :math:`f_{ZZ}` can be described as the difference between the frequency
        of qubit 0 when qubit 1 is excited and the frequency of qubit 0 when
        qubit 1 is in the ground state. Because :math:`f_{ZZ}` is symmetric in
        qubit index, it can also be expressed with the roles of 0 and 1
        reversed.  Experimentally, we measure :math:`f_{ZZ}` by performing
        Ramsey sequences on qubit 0 with qubit 1 in the ground state and again
        with qubit 1 in the excited state. The standard Ramsey experiment
        consists of putting a qubit along the :math:`X` axis of Bloch sphere,
        waiting for some time, and then measuring the qubit project along
        :math:`X`. By measuring the :math:`X` projection versus time the qubit
        frequency can be inferred. See
        :class:`~qiskit_experiments.library.characterization.T2Ramsey` and
        :class:`~qiskit_experiments.library.characterization.RamseyXY`.

        Because we are interested in the difference in qubit 0 frequency
        between the two qubit 1 preparations rather than the absolute
        frequencies of qubit 0 for those preparations, we modify the Ramsey
        sequences (the circuits for the modified sequences are shown below).
        First, we add an X gate on qubit 0 to the middle of the Ramsey delay.
        This would have the effect of echoing out the phase accumulation of
        qubit 0 (like a Hahn echo sequence as used in
        :class:`~qiskit_experiments.library.characterization.T2Hahn`), but we
        add a simultaneous X gate to qubit 1 as well.  Flipping qubit 1 inverts
        the sign of the :math:`f_{ZZ}` term. The net result is that qubit 0
        continues to accumulate phase proportional to :math:`f_{ZZ}` while the
        phase due to any ZI term is canceled out. This technique allows
        :math:`f_{ZZ}` to be measured using longer delay times than might
        otherwise be possible with a qubit with a slow frequency drift (i.e.
        the measurement is not sensitive to qubit frequency drift from shot to
        shot, only to drift within a single shot).

        The resulting excited state population of qubit 0 versus delay time
        exhibits slow sinusoidal oscillations (assuming :math:`f_{ZZ}` is
        relatively small). To help with distinguishing between qubit decay and
        a slow oscillation, an extra Z rotation is applied before the final
        pulse on qubit 0. The angle of this Z rotation is set proportional to
        the delay time of the sequence. This angle proportional to time behaves
        similarly to measuring at a fixed angle with the qubit rotating at a
        constant frequency. This virtual frequency is common to the two qubit 1
        preparations. By looking at the difference in frequency fitted for the
        two cases, this virtual frequency (called :math:`f` in the circuits
        shown below) is removed, leaving only the :math:`f_{ZZ}` value. The
        value of :math:`f` in terms of the experiment options is
        ``num_rotations / (max(delays) - min(delays))``.

        This experiment consists of the following two circuits repeated with
        different ``delay`` values.

        .. parsed-literal::

            Modified Ramsey sequence with qubit 1 initially in the ground state

                 ┌────┐ ░ ┌─────────────────┐ ░ ┌───┐ ░ ┌─────────────────┐ ░ »
            q_0: ┤ √X ├─░─┤ Delay(delay[s]) ├─░─┤ X ├─░─┤ Delay(delay[s]) ├─░─»
                 └────┘ ░ └─────────────────┘ ░ ├───┤ ░ └─────────────────┘ ░ »
            q_1: ───────░─────────────────────░─┤ X ├─░─────────────────────░─»
                        ░                     ░ └───┘ ░                     ░ »
            c: 1/═════════════════════════════════════════════════════════════»
                                                                              »
            «     ┌─────────────────────┐┌────┐ ░ ┌─┐
            «q_0: ┤ Rz(4*delay*dt*f*pi) ├┤ √X ├─░─┤M├
            «     └────────┬───┬────────┘└────┘ ░ └╥┘
            «q_1: ─────────┤ X ├────────────────░──╫─
            «              └───┘                ░  ║
            «c: 1/═════════════════════════════════╩═
            «                                      0

            Modified Ramsey sequence with qubit 1 initially in the excited state

                 ┌────┐ ░ ┌─────────────────┐ ░ ┌───┐ ░ ┌─────────────────┐ ░ »
            q_0: ┤ √X ├─░─┤ Delay(delay[s]) ├─░─┤ X ├─░─┤ Delay(delay[s]) ├─░─»
                 ├───┬┘ ░ └─────────────────┘ ░ ├───┤ ░ └─────────────────┘ ░ »
            q_1: ┤ X ├──░─────────────────────░─┤ X ├─░─────────────────────░─»
                 └───┘  ░                     ░ └───┘ ░                     ░ »
            c: 1/═════════════════════════════════════════════════════════════»
                                                                              »
            «     ┌─────────────────────┐┌────┐ ░ ┌─┐
            «q_0: ┤ Rz(4*delay*dt*f*pi) ├┤ √X ├─░─┤M├
            «     └─────────────────────┘└────┘ ░ └╥┘
            «q_1: ──────────────────────────────░──╫─
            «                                   ░  ║
            «c: 1/═════════════════════════════════╩═
            «                                      0

    # section: analysis_ref

        :class:`ZZRamseyAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
            from qiskit_experiments.test.patching import patch_sampler_test_support
            patch_sampler_test_support()

            # backend
            from qiskit_ibm_runtime.fake_provider import FakePerth
            from qiskit_aer import AerSimulator
            from qiskit_aer.noise import NoiseModel

            noise_model = NoiseModel.from_backend(FakePerth(),
                                                  thermal_relaxation=True,
                                                  gate_error=False,
                                                  readout_error=False,
            )

            backend = AerSimulator.from_backend(FakePerth(), noise_model=noise_model)

        .. jupyter-execute::

            from qiskit_experiments.library.characterization import ZZRamsey

            exp = ZZRamsey(physical_qubits=(0,1), backend=backend)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    def __init__(
        self,
        physical_qubits: Tuple[int, int],
        backend: Union[Backend, None] = None,
        **experiment_options,
    ):
        """Create new experiment.

        Args:
            physical_qubits: The qubits on which to run the Ramsey XY experiment.
            backend: Optional, the backend to run the experiment on.
            experiment_options: experiment options to set
        """
        super().__init__(physical_qubits, analysis=ZZRamseyAnalysis(), backend=backend)
        # Override the default of get_processor() which is "1" * num_qubits. We
        # only fit the probability of the target qubit.
        self.analysis.set_options(outcome="1")
        self.set_experiment_options(**experiment_options)

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default values for the :math:`ZZ` Ramsey experiment.

        Experiment Options:
            delays (list[float]): The list of delays that will be scanned in
                the experiment, in seconds. If not set, then ``num_delays``
                evenly spaced delays between ``min_delay`` and ``max_delay``
                are used. If ``delays`` is set, ``max_delay``, ``min_delay``,
                and ``num_delays`` are ignored.
            max_delay (float): Maximum delay time to use.
            min_delay (float): Minimum delay time to use.
            num_delays (int): Number of circuits to use per control state
                preparation.
            num_rotations (float): The extra rotation added to qubit 0 uses a
                frequency that gives this many rotations in the case where
                :math:`f_{ZZ}` is 0.
        """
        options = super()._default_experiment_options()
        options.delays = None
        options.min_delay = 0e-6
        options.max_delay = 10e-6
        options.num_delays = 50
        options.num_rotations = 5

        return options

    def delays(self) -> List[float]:
        """Delay values to use in circuits

        Returns:
            The list of delays to use for the different circuits based on the
            experiment options.
        """
        # This method allows delays to be set by the user as an explicit
        # sequence or as a minimum, maximum and number of values for a linearly
        # spaced sequence.
        options = self.experiment_options
        if options.delays is not None:
            return options.delays

        return np.linspace(options.min_delay, options.max_delay, options.num_delays).tolist()

    def frequency(self) -> float:
        """Frequency of qubit rotation when ZZ is 0

        This method calculates the simulated frequency applied to both sets of
        circuits. The value is chosen to induce `num_rotations` number of
        rotation within the time window that the delay is swept through.

        Returns:
            The frequency at which the target qubit will rotate when ZZ is zero
            based on the current experiment options.
        """
        delays = self.delays()
        freq = self.experiment_options.num_rotations / (max(delays) - min(delays))

        return freq

    def _template_circuits(
        self,
        frequency: Union[None, float, ParameterExpression] = None,
    ) -> Tuple[QuantumCircuit, QuantumCircuit]:
        """Template circuits for series 0 and 1 parameterized by delay

        The generated circuits have the length of the delay instructions as the
        only parameter.

        Args:
            dt_value: the value of the backend ``dt`` value. Used to convert
                delay values into units of seconds.
            delay_unit: the unit of circuit delay instructions.

        Returns:
            Circuits for series 0 and 1
        """
        delay = Parameter("delay")

        timing = BackendTiming(self.backend)

        frequency = frequency if frequency is not None else self.frequency()
        # frequency is always in units of Hz.  delay_freq has inverse units to
        # the units of `delay`.
        #
        # If the backend does not have a `dt`, the delays will be treated as in
        # units of seconds. Otherwise they will be in units of samples. For the
        # samples case, we multiply by `dt` so that `delay_freq` is in inverse
        # samples per cycle.
        if timing.delay_unit != "s":
            delay_freq = timing.dt * frequency
        else:
            delay_freq = frequency

        # Template circuit for series 0
        # Control qubit starting in |0> state, flipping to |1> in middle
        circ0 = QuantumCircuit(2, 1)
        circ0.metadata["series"] = "0"

        circ0.sx(0)

        circ0.barrier()
        circ0.delay(delay, 0, timing.delay_unit)
        circ0.barrier()

        circ0.x(0)
        circ0.x(1)

        circ0.barrier()
        circ0.delay(delay, 0, timing.delay_unit)
        circ0.barrier()

        circ0.rz(2 * np.pi * delay_freq * (2 * delay), 0)
        circ0.sx(0)
        # Flip control back to 0, so control qubit is in 0 for all circuits
        # when qubit 1 is measured.
        circ0.x(1)

        circ0.barrier()
        circ0.measure(0, 0)

        # Template circuit for series 1
        # Control qubit starting in |1> state, flipping to |0> in middle
        circ1 = QuantumCircuit(2, 1)
        circ1.metadata["series"] = "1"

        circ1.x(1)
        circ1.sx(0)

        circ1.barrier()
        circ1.delay(delay, 0, timing.delay_unit)
        circ1.barrier()

        circ1.x(0)
        circ1.x(1)

        circ1.barrier()
        circ1.delay(delay, 0, timing.delay_unit)
        circ1.barrier()

        circ1.rz(2 * np.pi * delay_freq * (2 * delay), 0)
        circ1.sx(0)

        circ1.barrier()
        circ1.measure(0, 0)

        return circ0, circ1

    def parametrized_circuits(self) -> Tuple[QuantumCircuit, QuantumCircuit]:
        r"""Create circuits with parameters for numerical quantities

        This method is primarily intended for generating template circuits that
        visualize well. It inserts :class:`qiskit.circuit.Parameter`\ s for
        :math:`π` and `dt` as well the target qubit rotation frequency `f`.

        Return:
            Parameterized circuits for the case of the control qubit being in 0
            and in 1.
        """
        f_param = Parameter("f")
        dt = Parameter("dt")
        pi = Parameter("pi")

        freq = dt * pi * f_param / np.pi

        timing = BackendTiming(self.backend)
        if timing.dt is not None:
            freq = freq / timing.dt

        circs = self._template_circuits(frequency=freq)

        return circs

    def circuits(self) -> List[QuantumCircuit]:
        """Create circuits

        Returns:
            A list of circuits with a variable delay.
        """
        timing = BackendTiming(self.backend)

        circ0, circ1 = self._template_circuits()
        circs = []

        for delay in self.delays():
            for circ in (circ0, circ1):
                assigned = circ.assign_parameters(
                    {circ.parameters[0]: timing.round_delay(time=delay / 2)}, inplace=False
                )
                assigned.metadata["xval"] = 2 * timing.delay_time(time=delay / 2)
                circs.append(assigned)

        return circs



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Quantum state tomography analysis
"""
from qiskit_experiments.framework import Options
from .basis import PauliMeasurementBasis
from .tomography_analysis import TomographyAnalysis


class StateTomographyAnalysis(TomographyAnalysis):
    """State tomography experiment analysis.

    # section: overview
        Fitter Functions

        Built-in fitter functions may be selected using the following string
        labels, refer to the corresponding functions documentation for additional
        details on the fitters.

        * ``"linear_inversion"``:
          :func:`~qiskit_experiments.library.tomography.fitters.linear_inversion` (Default)
        * ``"cvxpy_linear_lstsq"``:
          :func:`~qiskit_experiments.library.tomography.fitters.cvxpy_linear_lstsq`
        * ``"cvxpy_gaussian_lstsq"``:
          :func:`~qiskit_experiments.library.tomography.fitters.cvxpy_gaussian_lstsq`

        PSD Rescaling

        For fitters that do not constrain the reconstructed state to be
        `positive-semidefinite` (PSD) we construct the maximum-likelihood
        nearest PSD state under the assumption of Gaussian measurement noise
        using the rescaling method in Reference [1]. For fitters that already
        support PSD constraints this option can be disabled by setting
        ``rescale_positive=False``.

    # section: warning
        The API for tomography fitters is still under development so may change
        in future releases.

    # section: note
        Fitters starting with ``"cvxpy_*"`` require the optional CVXPY Python
        package to be installed.

    # section: reference
        .. ref_arxiv:: 1 1106.5458

    """

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options

        Analysis Options:
            measurement_basis (:class:`~qiskit_experiments.library.tomography.basis.MeasurementBasis`):
                The measurement
                :class:`~qiskit_experiments.library.tomography.basis.MeasurementBasis`
                to use for tomographic state reconstruction.
            fitter (str or Callable): The fitter function to use for reconstruction.
                This can be a string to select one of the built-in fitters, or a callable to
                supply a custom fitter function. See the `Fitter Functions` section for
                additional information.
            fitter_options (dict): Any addition kwarg options to be supplied to the fitter
                function. For documentation of available kwargs refer to the fitter function
                documentation.
            rescale_positive (bool): If True rescale the state returned by the fitter
                to be positive-semidefinite. See the `PSD Rescaling` section for
                additional information (Default: True).
            rescale_trace (bool): If True rescale the state returned by the fitter
                have either trace 1 for :class:`~qiskit.quantum_info.DensityMatrix`,
                or trace dim for :class:`~qiskit.quantum_info.Choi` matrices (Default: True).
            measurement_qubits (Sequence[int]): Optional, the physical qubits with tomographic
                measurements. If not specified will be set to ``[0, ..., N-1]`` for N-qubit
                tomographic measurements.
            target (str or :class:`~qiskit.quantum_info.DensityMatrix`
                or :class:`~qiskit.quantum_info.Statevector`): Optional,
                set a custom target quantum state for computing the
                :func:`~qiskit.quantum_info.state_fidelity`
                of the fitted state against (Default: None).
            conditional_circuit_clbits (list[int]): Optional, the clbit indices in the
                source circuit to be conditioned on when reconstructing the state.
                Enabling this will return a list of reconstructed state components
                conditional on the values of these clbit values. The integer value of the
                conditioning clbits is stored in state analysis result extra field
                `"conditional_circuit_outcome"`.
            conditional_measurement_indices (list[int]): Optional, indices of tomography
                measurement qubits to used for conditional state reconstruction. Enabling
                this will return a list of reconstructed state components conditioned on
                the remaining tomographic bases conditional on the basis index, and outcome
                value for these measurements. The conditional measurement basis index and
                integer value of the measurement outcome is stored in state analysis result
                extra fields `"conditional_measurement_index"` and
                `"conditional_measurement_outcome"` respectively.
        """
        options = super()._default_options()
        options.measurement_basis = PauliMeasurementBasis()
        return options



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Entry for artifact data.
"""

from dataclasses import dataclass, field
from typing import Any, Optional, List
from datetime import datetime
import uuid

from dateutil import tz


@dataclass
class ArtifactData:
    """A dataclass for non-analysis result payloads in :class:`.ExperimentData` objects.

    This class can convert results generated by the analysis class into a payload
    for saving and retrieving to and from the experiments service, which stores this
    as artifacts. Types of objects that may be converted to artifacts include fitted and raw data,
    fit status, and any other JSON-based data needed to serialize experiments and experiment data.

    Attributes:
        name: The name of the artifact. When saved to the cloud service, this will be the name
            of the zipfile this artifact object is stored in.
        data: The artifact payload.
        artifact_id: Artifact ID. Must be unique inside an :class:`ExperimentData` object.
        experiment_id: Experiment ID that the artifact is associated with.
        experiment: The name of the experiment.
        device_components: The device components of the experiment.
        created_time: Time when the artifact was created.
    """

    name: str
    data: Any
    artifact_id: Optional[str] = field(default_factory=lambda: str(uuid.uuid4()))
    experiment_id: Optional[str] = None
    experiment: Optional[str] = None
    device_components: List = field(default_factory=list)
    created_time: Optional[datetime] = field(default_factory=lambda: datetime.now(tz.tzlocal()))

    @property
    def dtype(self):
        """Data type of the payload."""
        return self.data.__class__.__name__

    def __repr__(self):
        return (
            f"ArtifactData(name={self.name}, dtype={self.dtype}, uid={self.artifact_id}, "
            f"experiment={self.experiment}, device_components={self.device_components})"
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Base plotter abstract class"""

import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

from qiskit_experiments.framework import Options
from qiskit_experiments.visualization.drawers import BaseDrawer, SeriesName

from ..style import PlotStyle


class BasePlotter(ABC):
    """An abstract class for the serializable figure plotters.

    # section: overview

    A plotter takes data from an experiment analysis class or experiment and plots a
    given figure using a drawing backend. Sub-classes define the kind of figure created
    and the expected data.

    Data is split into series and supplementary data. Series data is grouped by series
    name (``Union[str, int, float]``). For :class:`CurveAnalysis`, this is the model
    name for a curve fit. For series data associated with a single series name and
    supplementary data, data values are identified by a data key (str). Different data
    per series and figure must have a different data key to avoid overwriting values.
    Experiment and analysis results can be passed to the plotter so appropriate graphics
    can be drawn on the figure canvas. Series data is added to the plotter using
    :meth:`set_series_data` whereas supplementary data is added using
    :meth:`set_supplementary_data`. Series and supplementary data are retrieved using
    :meth:`data_for` and :attr:`supplementary_data` respectively.

    Series data contains values to be plotted on a canvas, such that the data can be
    grouped into subsets identified by their series name. Series names can be thought of
    as legend labels for the plotted data, and as curve names for a curve fit.
    Supplementary data is not associated with a series or curve and is instead only
    associated with the figure. Examples include analysis reports or other text that is
    drawn onto the figure canvas.

    .. rubric:: Options and Figure Options

    Plotters have both :attr:`options` and :attr:`figure_options` available to set
    parameters that define how to plot and what is plotted. :class:`BaseDrawer` is
    similar in that it also has ``options`` and ``figure_options``. The former contains
    class-specific variables that define how an instance behaves. The latter contains
    figure-specific variables that typically contain values that are drawn on the
    canvas, such as text.

    For example, :class:`BasePlotter` has an ``axis`` option that can be set to the
    canvas on which the figure should be drawn. This changes how the plotter works in
    that it changes where the figure is drawn. :class:`BasePlotter` has an ``xlabel``
    figure option that can be set to change the text drawn next to the X-axis in the
    final figure. As the value of this option will be drawn on the figure, it is a
    figure option.

    As plotters need a drawer to generate a figure, and the drawer needs to know what to
    draw, figure options are passed to :attr:`drawer` when the :meth:`figure` method is
    called. Any figure options that are defined in both the plotters
    :attr:`figure_options` attribute and the drawers ``figure_options`` attribute are
    copied to the drawer: i.e., :meth:`BaseDrawer.set_figure_options` is called for each
    common figure option, setting the value of the option to the value stored in the
    plotter.

    .. note::
        If a figure option called "foo" is not set in the drawer's figure options
        (:attr:`BaseDrawer.figure_options`) but is set in the plotter's figure options
        (:attr:`figure_options`), it will not be copied over to the drawer when the
        :meth:`figure` method is called. This means that some figure options from the
        plotter may be unused by the drawer. :class:`BasePlotter` and its subclasses
        filter these options before setting them in the drawer, as subclasses of
        :class:`BaseDrawer` may add additional figure options. To make validation
        easier and the code cleaner, the :meth:`figure` method conducts this check
        before setting figure options in the drawer.

    .. rubric:: Example

    .. code-block:: python

        plotter = MyPlotter(MyDrawer())

        # MyDrawer contains the following figure_options with default values.
        plotter.drawer.figure_options.xlabel
        plotter.drawer.figure_options.ylabel

        # MyDrawer does NOT contain the following figure option
        # plotter.drawer.figure_options.unknown_variable    # Raises an error as it
                                                            # does not exist in
                                                            # `drawer.figure_options`.

        # If we set the following figure options, they will be set in the drawer as
        # they are defined in `plotter.drawer.figure_options`.
        plotter.set_figure_options(xlabel="Frequency", ylabel="Fidelity")

        # During a call to `plotter.figure()`, the drawer's figure options are updated.
        # The following values would then be returned from the drawer.
        plotter.drawer.figure_options.xlabel                # returns "Frequency"
        plotter.drawer.figure_options.ylabel                # returns "Fidelity"

        # If we set the following option and figure option, they will NOT be set in the
        # drawer as the drawer doesn't contain default values for these option names.
        plotter.set_options(plot_fit=False)                 # Example plotter option
        plotter.set_figure_options(unknown_variable=5e9)    # Example figure option

        # As `plot_fit` is not a figure option, it is not set in the drawer.
        plotter.drawer.options.plot_fit     # Would raise an error if no default
                                            # exists, or return a different value to
                                            # `plotter.options.plot_fit`.

        # As `unknown_variable` is not set in the drawer's figure options, it is not set
        # during a # call to the `figure()` method.
        # plotter.drawer.figure_options.unknown_variable    # Raises an error as it
                                                            # does not exist in
                                                            # `drawer.figure_options`.

    Attributes:
        drawer (BaseDrawer): The drawer to use when plotting.
    """

    def __init__(self, drawer: BaseDrawer):
        """Create a new plotter instance.

        Args:
            drawer: The drawer to use when creating the figure.
        """
        # Data to be plotted, such as scatter points, interpolated fits, and confidence intervals
        self._series_data: Dict[SeriesName, Dict[str, Any]] = {}
        # Data that isn't directly associated with a single series, such as text or fit reports.
        self._supplementary_data: Dict[str, Any] = {}

        # Options for the plotter
        self._options = self._default_options()
        # Plotter options that have changed, for serialization.
        self._set_options = set()

        # Figure options that are updated in the drawer when `plotter.figure()` is called
        self._figure_options = self._default_figure_options()
        # Figure options that have changed, for serialization.
        self._set_figure_options = set()

        self.drawer: BaseDrawer = drawer

    @property
    def supplementary_data(self) -> Dict[str, Any]:
        """Additional data for the figure being plotted, that isn't associated with a
        series.

        Supplementary data includes text, fit reports, or other data that is associated
        with the figure but not an individual series. It is typically data additional to
        the direct results of an experiment.
        """
        return self._supplementary_data

    @property
    def series_data(self) -> Dict[SeriesName, Dict[str, Any]]:
        """Data for series being plotted.

        Series data includes data such as scatter points, interpolated fit values, and
        standard deviations. Series data is grouped by series name (``Union[str, int,
        float]``) and then by a data key (``str``). Though series data can be accessed
        through :attr:`series_data`, it is recommended to access them with
        :meth:`data_for` and :meth:`data_exists_for` as they allow for easier access to
        nested values and can handle multiple data keys in one query.

        Returns:
            A dictionary containing series data.
        """
        return self._series_data

    @property
    def series(self) -> List[SeriesName]:
        """Series names that have been added to this plotter."""
        return list(self._series_data.keys())

    def data_keys_for(self, series_name: SeriesName) -> List[str]:
        """Returns a list of data keys for the given series.

        Args:
            series_name: The series name for which to return the data keys, i.e., the
                types of data for each series.

        Returns:
            The list of data keys for data in the plotter associated with the given
            series. If the series has not been added to the plotter, an empty list is
            returned.
        """
        return list(self._series_data.get(series_name, []))

    def data_for(
        self, series_name: SeriesName, data_keys: Union[str, List[str]]
    ) -> Tuple[Optional[Any]]:
        """Returns data associated with the given series.

        The returned tuple contains the data, associated with ``data_keys``, in the same
        orders as they are provided. For example,

        .. code-block:: python

            plotter.set_series_data("seriesA", x=data.x, y=data.y, yerr=data.yerr)

            # The following calls are equivalent.
            x, y, yerr = plotter.data_for("seriesA", ["x", "y", "yerr"])
            x, y, yerr = data.x, data.y, data.yerr

            # Retrieving a single data key returns a tuple. Note the comma after ``x``.
            x, = plotter.data_for("seriesA", "x")

        :meth:`data_for` is intended to be used by sub-classes of :class:`BasePlotter`
        when plotting in the :meth:`_plot_figure` method.

        Args:
            series_name: The series name for the given series.
            data_keys: List of data keys for the data to be returned. If a single
                data key is given as a string, it is wrapped in a list.

        Returns:
            A tuple of data associated with the given series, identified by
            ``data_keys``. If no data has been set for a data key, None is returned for
            the associated tuple entry.
        """

        # We may be given a single data key, but we need a list for the rest of the function.
        if not isinstance(data_keys, list):
            data_keys = [data_keys]

        # The series doesn't exist in the plotter data, return None for each data key in the output.
        if series_name not in self._series_data:
            return (None,) * len(data_keys)

        return tuple(self._series_data[series_name].get(key, None) for key in data_keys)

    def set_series_data(self, series_name: SeriesName, **data_kwargs):
        """Sets data for the given series.

        Note that if data has already been assigned for the given series and data key,
        it will be overwritten with the new values. ``set_series_data`` will warn if the
        data key is unexpected; i.e., not within those returned by
        :meth:`expected_series_data_keys`.

        Args:
            series_name: The name of the given series.
            data_kwargs: The data to be added, where the keyword is the data key.
        """
        # Warn if the data keys are not expected.
        unknown_data_keys = [
            data_key for data_key in data_kwargs if data_key not in self.expected_series_data_keys()
        ]
        for unknown_data_key in unknown_data_keys:
            warnings.warn(
                f"{self.__class__.__name__} encountered an unknown data key {unknown_data_key}. It may "
                "not be used by the plotter class."
            )

        # Set data
        if series_name not in self._series_data:
            self._series_data[series_name] = {}
        self._series_data[series_name].update(**data_kwargs)

    def clear_series_data(self, series_name: Optional[SeriesName] = None):
        """Clear series data for this plotter.

        Args:
            series_name: The series name identifying which data should be cleared. If
                None, all series data is cleared. Defaults to None.
        """
        if series_name is None:
            self._series_data = {}
        elif series_name in self._series_data:
            self._series_data.pop(series_name)

    def set_supplementary_data(self, **data_kwargs):
        """Sets supplementary data for the plotter.

        Supplementary data differs from series data in that it is not associate with a
        series name. Fit reports are examples of supplementary data as they contain fit
        results from an analysis class, such as the "goodness" of a curve fit.

        Note that if data has already been assigned for the given data key, it will be
        overwritten with the new values. ``set_supplementary_data`` will warn if the
        data key is unexpected; i.e., not within those returned by
        :meth:`expected_supplementary_data_keys`.

        """

        # Warn if any data keys are not expected.
        unknown_data_keys = [
            data_key
            for data_key in data_kwargs
            if data_key not in self.expected_supplementary_data_keys()
        ]
        for unknown_data_key in unknown_data_keys:
            warnings.warn(
                f"{self.__class__.__name__} encountered an unknown data key {unknown_data_key}. It may "
                "not be used by the plotter class."
            )

        self._supplementary_data.update(**data_kwargs)

    def clear_supplementary_data(self):
        """Clears supplementary data."""
        self._supplementary_data = {}

    def data_exists_for(self, series_name: SeriesName, data_keys: Union[str, List[str]]) -> bool:
        """Returns whether the given data keys exist for the given series.

        Args:
            series_name: The name of the given series.
            data_keys: The data keys to be checked.

        Returns:
            True if all data keys have values assigned for the given series. False if at
            least one does not have a value assigned.
        """
        if not isinstance(data_keys, list):
            data_keys = [data_keys]

        # Handle non-existent series name
        if series_name not in self._series_data:
            return False

        return all(key in self._series_data[series_name] for key in data_keys)

    @abstractmethod
    def _plot_figure(self):
        """Generates a figure using :attr:`drawer` and :meth:`data`.

        Sub-classes must override this function to plot data using the drawer. This
        function is called by :meth:`figure` when :attr:`drawer` can be used to draw on
        the canvas.
        """

    def figure(self) -> Any:
        """Generates and returns a figure for the already provided series and
        supplementary data.

        :meth:`figure` calls :meth:`_plot_figure`, which is overridden by sub-classes.
        Before and after calling :meth:`_plot_figure`;
        :meth:`~BaseDrawer._configure_drawer`, :meth:`~BaseDrawer.initialize_canvas` and
        :meth:`~BaseDrawer.format_canvas` are called on the drawer respectively.

        Returns:
            A figure generated by :attr:`drawer`, of the same type as ``drawer.figure``.
        """
        # Initialize drawer, to copy axis, subplots, style, and figure options across.
        self._configure_drawer()

        # Initialize canvas, which creates subplots, assigns axis labels, etc.
        self.drawer.initialize_canvas()

        # Plot figure for given subclass. This is the core of BasePlotter subclasses.
        self._plot_figure()

        # Final formatting of canvas, which sets axis limits etc.
        self.drawer.format_canvas()

        # Return whatever figure is created by the drawer.
        return self.drawer.figure

    @classmethod
    @abstractmethod
    def expected_series_data_keys(cls) -> List[str]:
        """Returns the expected series data keys supported by this plotter."""

    @classmethod
    @abstractmethod
    def expected_supplementary_data_keys(cls) -> List[str]:
        """Returns the expected supplementary data keys supported by this plotter."""

    @property
    def options(self) -> Options:
        """Options for the plotter.

        Options for a plotter modify how the class generates a figure. This includes an
        optional axis object, being the drawer canvas. Make sure verify whether the
        option you want to set is in :attr:`options` or :attr:`figure_options`.
        """
        return self._options

    @property
    def figure_options(self) -> Options:
        """Figure options for the plotter and its drawer.

        Figure options differ from normal options (:attr:`options`) in that the plotter
        passes figure options on to the drawer when creating a figure (when
        :meth:`figure` is called). This way :attr:`drawer` can draw an appropriate
        figure. An example of a figure option is the x-axis label.
        """
        return self._figure_options

    @classmethod
    def _default_options(cls) -> Options:
        """Return default plotter options.

        Options:
            axis (Any): Arbitrary object that can be used as a drawing canvas.
            subplots (Tuple[int, int]): Number of rows and columns when the experimental
                result is drawn in the multiple windows.
            style (PlotStyle): The style definition to use when plotting.
                This overwrites figure option `custom_style` set in :attr:`drawer`. The
                default is an empty style object, and such the default :attr:`drawer`
                plotting style will be used.
        """
        return Options(
            axis=None,
            subplots=(1, 1),
            style=PlotStyle(),
        )

    @classmethod
    def _default_figure_options(cls) -> Options:
        """Return default figure options.

        Figure Options:
            xlabel (Union[str, List[str]]): X-axis label string of the output figure. If
                there are multiple columns in the canvas, this could be a list of labels.
            ylabel (Union[str, List[str]]): Y-axis label string of the output figure. If
                there are multiple rows in the canvas, this could be a list of labels.
            xlim (Union[Tuple[float, float], List[Tuple[float, float]]): Min and max value
                of the horizontal axis. If not provided, it is automatically scaled based
                on the input data points. If there are multiple columns in the canvas,
                this could be a list of xlims.
            ylim (Union[Tuple[float, float], List[Tuple[float, float]]): Min and max value
                of the vertical axis. If not provided, it is automatically scaled based
                on the input data points. If there are multiple rows in the canvas,
                this could be a list of ylims.
            xval_unit (Union[str, List[str]]): Unit of x values.
                No scaling prefix is needed here as this is controlled by ``xval_unit_scale``.
                If there are multiple columns in the canvas, this could be a list of xval_units.
            yval_unit (Union[str, List[str]]): Unit of y values.
                No scaling prefix is needed here as this is controlled by ``yval_unit_scale``.
                If there are multiple rows in the canvas, this could be a list of yval_units.
            xval_unit_scale (Union[bool, List[bool]]): Whether to add an SI unit prefix to
                ``xval_unit`` if needed. For example, when the x values represent time and
                ``xval_unit="s"``, ``xval_unit_scale=True`` adds an SI unit prefix to
                ``"s"`` based on X values of plotted data. In the output figure, the
                prefix is automatically selected based on the maximum value in this
                axis. If your x values are in [1e-3, 1e-4], they are displayed as [1 ms,
                10 ms]. By default, this option is set to ``True``. If ``False`` is
                provided, the axis numbers will be displayed in the scientific notation.
                If there are multiple columns in the canvas, this could be a list of xval_unit_scale.
            yval_unit_scale (Union[bool, List[bool]]): Whether to add an SI unit prefix to
                ``yval_unit`` if needed. See ``xval_unit_scale`` for details.
                If there are multiple rows in the canvas, this could be a list of yval_unit_scale.
            xscale (str): The scaling of the x-axis, such as ``log`` or ``linear``.
            yscale (str): The scaling of the y-axis, such as ``log`` or ``linear``.
            figure_title (str): Title of the figure. Defaults to None, i.e. nothing is
                shown.
            sharex (bool): Set True to share x-axis ticks among sub-plots.
            sharey (bool): Set True to share y-axis ticks among sub-plots.
            series_params (Dict[str, Dict[str, Any]]): A dictionary of parameters for
                each series. This is keyed on the name for each series. Sub-dictionary
                is expected to have the following three configurations, "canvas",
                "color", "symbol" and "label"; "canvas" is the integer index of axis
                (when multi-canvas plot is set), "color" is the color of the drawn
                graphics, "symbol" is the series marker style for scatter plots, and
                "label" is a user provided series label that appears in the legend.
        """
        options = Options(
            xlabel=None,
            ylabel=None,
            xlim=None,
            ylim=None,
            xval_unit=None,
            yval_unit=None,
            xval_unit_scale=True,
            yval_unit_scale=True,
            xscale=None,
            yscale=None,
            sharex=True,
            sharey=True,
            figure_title=None,
            series_params={},
        )
        options.set_validator("xscale", ["linear", "log", "symlog", "logit", "quadratic", None])
        options.set_validator("yscale", ["linear", "log", "symlog", "logit", "quadratic", None])

        return options

    def set_options(self, **fields):
        """Set the plotter options.

        Args:
            fields: The fields to update in options.

        Raises:
            AttributeError: If an unknown option is encountered.
        """
        for field in fields:
            if not hasattr(self._options, field):
                raise AttributeError(
                    f"Options field {field} is not valid for {type(self).__name__}"
                )
        self._options.update_options(**fields)
        self._set_options = self._set_options.union(fields)

    def set_figure_options(self, **fields):
        """Set the figure options.

        Args:
            fields: The fields to update in figure options.
        """
        # Don't check if any option in fields already exists (like with `set_options`), as figure options
        # are passed to `.drawer` which may have other figure options. Any figure option that isn't set
        # in `.drawer.figure_options` won't be set anyway. Setting `.drawer.figure_options` only occurs
        # in `.figure()`, so we can't compare to `.drawer.figure_options` now as `.drawer` may be changed
        # between now and the call to `.figure()`.
        self._figure_options.update_options(**fields)
        self._set_figure_options = self._set_figure_options.union(fields)

    def _configure_drawer(self):
        """Configures :attr:`drawer` before plotting.

        The following actions are taken:

        1. ``axis``, ``subplots``, and ``style`` are passed to :attr:`drawer`.
        2. ``figure_options`` in :attr:`drawer` are updated based on values set in
           the plotter :attr:`figure_options`

        These steps are different as all figure options could be passed to
        :attr:`drawer`, if the drawer already has a figure option with the same name.
        ``axis``, ``subplots``, and ``style`` are the only plotter options (from
        :attr:`options`) passed to :attr:`drawer` in :meth:`_configure_drawer`. This is
        done as these options make more sense as an option for a plotter, given the
        interface of :class:`BasePlotter`.
        """
        ## Axis, subplots, and style
        if self.options.axis:
            self.drawer.set_options(axis=self.options.axis)
        if self.options.subplots:
            self.drawer.set_options(subplots=self.options.subplots)
        self.drawer.set_figure_options(custom_style=self.options.style)

        # Convert options to dictionaries for easy comparison of all options/fields.
        _drawer_figure_options = self.drawer.figure_options.__dict__
        _plotter_figure_options = self.figure_options.__dict__

        # If an option exists in drawer.figure_options AND in self.figure_options, set the drawer's
        # figure option value to that from the plotter.
        for opt_key in _drawer_figure_options:
            if opt_key in _plotter_figure_options:
                _drawer_figure_options[opt_key] = _plotter_figure_options[opt_key]

        # Use drawer.set_figure_options so figure options are serialized.
        self.drawer.set_figure_options(**_drawer_figure_options)

    def config(self) -> Dict:
        """Return the config dictionary for this drawing."""
        options = dict((key, getattr(self._options, key)) for key in self._set_options)
        figure_options = dict(
            (key, getattr(self._figure_options, key)) for key in self._set_figure_options
        )
        drawer = self.drawer.__json_encode__()

        return {
            "cls": type(self),
            "options": options,
            "figure_options": figure_options,
            "drawer": drawer,
        }

    def __json_encode__(self):
        return self.config()

    @classmethod
    def __json_decode__(cls, value):
        ## Process drawer as it's needed to create a plotter
        drawer_values = value["drawer"]
        # We expect a subclass of BaseDrawer
        drawer_cls: BaseDrawer = drawer_values["cls"]
        drawer = drawer_cls.__json_decode__(drawer_values)

        # Create plotter instance
        instance = cls(drawer)
        if "options" in value:
            instance.set_options(**value["options"])
        if "figure_options" in value:
            instance.set_figure_options(**value["figure_options"])
        return instance



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""The analysis class for the Ramsey XY experiment."""

from typing import List, Union

import lmfit
import numpy as np

import qiskit_experiments.curve_analysis as curve


class RamseyXYAnalysis(curve.CurveAnalysis):
    r"""Ramsey XY analysis based on a fit to a cosine function and a sine function.

    # section: fit_model

        Analyze a Ramsey XY experiment by fitting the X and Y series to a cosine and sine
        function, respectively. The two functions share the frequency and amplitude parameters.

        .. math::

            y_X = {\rm amp}e^{-x/\tau}\cos\left(2\pi\cdot{\rm freq}_i\cdot x\right) + {\rm base} \\
            y_Y = {\rm amp}e^{-x/\tau}\sin\left(2\pi\cdot{\rm freq}_i\cdot x\right) + {\rm base}

    # section: fit_parameters
        defpar \rm amp:
            desc: Amplitude of both series.
            init_guess: Half of the maximum y value less the minimum y value. When the
                oscillation frequency is low, it uses an averaged difference of
                Ramsey X data - Ramsey Y data.
            bounds: [0, 2 * average y peak-to-peak]

        defpar \tau:
            desc: The exponential decay of the curve.
            init_guess: The initial guess is obtained by fitting an exponential to the
                square root of (X data)**2 + (Y data)**2.
            bounds: [0, inf]

        defpar \rm base:
            desc: Base line of both series.
            init_guess: Roughly the average of the data. When the oscillation frequency is low,
                it uses an averaged data of Ramsey Y experiment.
            bounds: [min y - average y peak-to-peak, max y + average y peak-to-peak]

        defpar \rm freq:
            desc: Frequency of both series. This is the parameter of interest.
            init_guess: The frequency with the highest power spectral density.
            bounds: [-inf, inf]

        defpar \rm phase:
            desc: Common phase offset.
            init_guess: 0
            bounds: [-pi, pi]
    """

    def __init__(self):
        super().__init__(
            models=[
                lmfit.models.ExpressionModel(
                    expr="amp * exp(-x / tau) * cos(2 * pi * freq * x + phase) + base",
                    name="X",
                ),
                lmfit.models.ExpressionModel(
                    expr="amp * exp(-x / tau) * sin(2 * pi * freq * x + phase) + base",
                    name="Y",
                ),
            ]
        )

    @classmethod
    def _default_options(cls):
        """Return the default analysis options.

        See :meth:`~qiskit_experiment.curve_analysis.CurveAnalysis._default_options` for
        descriptions of analysis options.
        """
        default_options = super()._default_options()
        default_options.data_subfit_map = {
            "X": {"series": "X"},
            "Y": {"series": "Y"},
        }
        default_options.plotter.set_figure_options(
            xlabel="Delay",
            ylabel="Signal (arb. units)",
            xval_unit="s",
        )
        default_options.result_parameters = ["freq"]

        return default_options

    def _generate_fit_guesses(
        self,
        user_opt: curve.FitOptions,
        curve_data: curve.ScatterTable,
    ) -> Union[curve.FitOptions, List[curve.FitOptions]]:
        """Create algorithmic initial fit guess from analysis options and curve data.

        Args:
            user_opt: Fit options filled with user provided guess and bounds.
            curve_data: Formatted data collection to fit.

        Returns:
            List of fit options that are passed to the fitter function.
        """
        ramx_data = curve_data.filter(series="X")
        ramy_data = curve_data.filter(series="Y")

        # At very low frequency, y value of X (Y) curve stay at P=1.0 (0.5) for all x values.
        # Computing y peak-to-peak with combined data gives fake amplitude of 0.25.
        # Same for base, i.e. P=0.75 is often estimated in this case.
        full_y_ptp = np.ptp(curve_data.y)
        avg_y_ptp = 0.5 * (np.ptp(ramx_data.y) + np.ptp(ramy_data.y))
        max_y = np.max(curve_data.y)
        min_y = np.min(curve_data.y)

        user_opt.bounds.set_if_empty(
            amp=(0, full_y_ptp * 2),
            tau=(0, np.inf),
            base=(min_y - avg_y_ptp, max_y + avg_y_ptp),
            phase=(-np.pi, np.pi),
        )

        if avg_y_ptp < 0.5 * full_y_ptp:
            # When X and Y curve don't oscillate, X (Y) usually stays at P(1) = 1.0 (0.5).
            # So peak-to-peak of full data is something around P(1) = 0.75, while
            # single curve peak-to-peak is almost zero.
            avg_x = np.average(ramx_data.y)
            avg_y = np.average(ramy_data.y)

            user_opt.p0.set_if_empty(
                amp=np.abs(avg_x - avg_y),
                tau=100 * np.max(curve_data.x),
                base=avg_y,
                phase=0.0,
                freq=0.0,
            )
            return user_opt

        base_guess_x = curve.guess.constant_sinusoidal_offset(ramx_data.y)
        base_guess_y = curve.guess.constant_sinusoidal_offset(ramy_data.y)
        base_guess = 0.5 * (base_guess_x + base_guess_y)
        user_opt.p0.set_if_empty(
            amp=0.5 * full_y_ptp,
            base=base_guess,
            phase=0.0,
        )

        # Guess the exponential decay by combining both curves
        ramx_unbiased = ramx_data.y - user_opt.p0["base"]
        ramy_unbiased = ramy_data.y - user_opt.p0["base"]
        decay_data = ramx_unbiased**2 + ramy_unbiased**2
        if np.ptp(decay_data) < 0.95 * 0.5 * full_y_ptp:
            # When decay is less than 95 % of peak-to-peak value, ignore decay and
            # set large enough tau value compared with the measured x range.
            user_opt.p0.set_if_empty(tau=1000 * np.max(curve_data.x))
        else:
            user_opt.p0.set_if_empty(tau=-1 / curve.guess.exp_decay(ramx_data.x, decay_data))

        # Guess the oscillation frequency, remove offset to eliminate DC peak
        freq_guess_x = curve.guess.frequency(ramx_data.x, ramx_unbiased)
        freq_guess_y = curve.guess.frequency(ramy_data.x, ramy_unbiased)
        freq_val = 0.5 * (freq_guess_x + freq_guess_y)

        # FFT might be up to 1/2 bin off
        df = 2 * np.pi / (np.min(np.diff(ramx_data.x)) * ramx_data.x.size)
        freq_guesses = [freq_val - df, freq_val + df, freq_val]

        # Ramsey XY is frequency sign sensitive.
        # Since experimental data is noisy, correct sign is hardly estimated with phase velocity.
        # Try both positive and negative frequency to find the best fit.
        opts = []
        for sign in (1, -1):
            for freq_guess in freq_guesses:
                opt = user_opt.copy()
                opt.p0.set_if_empty(freq=sign * freq_guess)
                opts.append(opt)

        return opts

    def _evaluate_quality(self, fit_data: curve.CurveFitResult) -> Union[str, None]:
        """Algorithmic criteria for whether the fit is good or bad.

        A good fit has:
            - a reduced chi-squared lower than three and greater than zero,
            - an error on the frequency smaller than the frequency.
        """
        fit_freq = fit_data.ufloat_params["freq"]

        criteria = [
            0 < fit_data.reduced_chisq < 3,
            curve.utils.is_error_not_significant(fit_freq),
        ]

        if all(criteria):
            return "good"

        return "bad"
